import os
from abc import ABC
from dataclasses import dataclass, field
from os import PathLike
from typing import List, Union, Any

__all__ = ['Path', 'Line', 'File', 'Directory']

from diagnose import fileutils


@dataclass
class Path(PathLike, ABC):
    _path: str

    def __fspath__(self) -> Any:
        return self._path


@dataclass
class Line:
    filename: str  # abs filename
    content: str
    lineno: int

    def __repr__(self):
        return f'<Line content={self.content} lineno={self.lineno} filename={self.filename}>'

    def __str__(self):
        return self.content

    def __hash__(self):
        """
        Generate the hash of the :class:`Line`.

        The hash can be used as the classname of the element for a line in code when rendered in the HTML file. Note
        that this magic method :meth:`__hash__` will return a string as a hash rather than an int.

        Returns:
            str: The hash of the :class:`line`.
        """

        # XXX: Hack to the megic method __hash__ and make it returns a string rather than an int.
        return fileutils.generate_classname(self.filename, self.lineno)


@dataclass
class File:
    filename: PathLike | str
    basename: str
    extension: str | None = field(default_factory=str)
    last_modified: str = field(default_factory=str)
    size: str = field(default_factory=str)
    lines: List['Line'] = field(default_factory=list)

    def __repr__(self):
        return f'<File filename={os.fspath(self.filename)} extension={self.extension}>'

    def __str__(self):
        return ''.join([line.content for line in self.lines])

    def __hash__(self):
        # XXX: Hack to the megic method __hash__ and make it returns a string rather than an int.
        return fileutils.generate_classname(self.filename)


@dataclass
class Directory:
    dirname: PathLike | str
    basename: str
    last_modified: str = field(default_factory=str)
    size: str = field(default_factory=str)
    files_or_directories: List[Union['File', 'Directory']] = field(default_factory=list)

    def __repr__(self):
        return f'<Directory dirname={os.fspath(self.dirname)}>'

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        # XXX: Hack to the megic method __hash__ and make it returns a string rather than an int.
        return fileutils.generate_classname(self.dirname)



from typing import Any

from diagnose.engine import PluginManager
from diagnose.logs import Logger
from diagnose.typings import T_event, T_frame, TPlugin


@PluginManager.add_plugin(enabled=False)
class RedirectPlugin(TPlugin):

    def __init__(self, *args, **kwargs):
        self.filename = None

    def set_filename(self, filename: str) -> None:
        self.filename = filename

    def on_preprocess(self):
        Logger.redirect_to_file(self.filename, logger=Logger.get_logger('engine'))

    def on_postprocess(self):
        pass

    def tracer_callback(self, frame: T_frame, event: T_event, args: Any) -> Any:
        pass


"""
Copyright (c) Aymeric Augustin and contributors

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import base64
import binascii
import enum
import hashlib
from typing import List

from diagnose.server.ws.exception import WebsocketException
from diagnose.server.ws.http11 import Headers
from diagnose.server.ws.parse import parse_upgrade, parse_connection
from diagnose.server.ws.typings import UpgradeProtocol, ConnectionOption


class State(enum.IntEnum):
    """A WebSocket connection is in one of these four states."""

    CONNECTING, OPEN, CLOSING, CLOSED = range(4)


GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"


def accept_key(key: str) -> str:
    """
    Compute the value of the Sec-WebSocket-Accept header.

    Args:
        key: value of the Sec-WebSocket-Key header.

    """
    sha1 = hashlib.sha1((key + GUID).encode()).digest()
    return base64.b64encode(sha1).decode()


def check_request(headers: Headers) -> str:
    """
    Check a handshake request received from the client.

    This function doesn't verify that the request is an HTTP/1.1 or higher GET
    request and doesn't perform ``Host`` and ``Origin`` checks. These controls
    are usually performed earlier in the HTTP request handling code. They're
    the responsibility of the caller.

    Args:
        headers: handshake request headers.

    Returns:
        str: ``key`` that must be passed to :func:`build_response`.

    Raises:
        InvalidHandshake: if the handshake request is invalid;
            then the server must return 400 Bad Request error.

    """
    connection: List[ConnectionOption] = sum(
        [parse_connection(value) for value in headers.get_all("Connection")], []
    )

    if not any(value.lower() == "upgrade" for value in connection):
        raise WebsocketException(f"InvalidUpgrade: Connection {','.join(connection)}")

    upgrade: List[UpgradeProtocol] = sum(
        [parse_upgrade(value) for value in headers.get_all("Upgrade")], []
    )

    # For compatibility with non-strict implementations, ignore case when
    # checking the Upgrade header. The RFC always uses "websocket", except
    # in section 11.2. (IANA registration) where it uses "WebSocket".
    if not (len(upgrade) == 1 and upgrade[0].lower() == "websocket"):
        raise WebsocketException(f"InvalidUpgrade: Upgrade {','.join(upgrade)}")

    try:
        s_w_key = headers["Sec-WebSocket-Key"]
    except KeyError as exc:
        raise WebsocketException("InvalidHeader: Sec-WebSocket-Key") from exc
    except Exception as exc:
        raise WebsocketException(
            "InvalidHeader: Sec-WebSocket-Key more than one Sec-WebSocket-Key header found"
        ) from exc

    try:
        raw_key = base64.b64decode(s_w_key.encode(), validate=True)
    except binascii.Error as exc:
        raise WebsocketException(f"InvalidHeaderValue: Sec-WebSocket-Key {s_w_key}") from exc
    if len(raw_key) != 16:
        raise WebsocketException(f"InvalidHeaderValue: Sec-WebSocket-Key {s_w_key}")

    try:
        s_w_version = headers["Sec-WebSocket-Version"]
    except KeyError as exc:
        raise WebsocketException("InvalidHeader: Sec-WebSocket-Version") from exc
    except Exception as exc:
        raise WebsocketException(
            "InvalidHeader: Sec-WebSocket-Version more than one Sec-WebSocket-Version header found"
        ) from exc

    if s_w_version != "13":
        raise WebsocketException(f"InvalidHeaderValue: Sec-WebSocket-Version {s_w_version}")

    return s_w_key


def build_response(headers: Headers, key: str) -> None:
    """
    Build a handshake response to send to the client.

    Update response headers passed in argument.

    Args:
        headers: handshake response headers.
        key: returned by :func:`check_request`.

    """
    headers["Upgrade"] = "websocket"
    headers["Connection"] = "Upgrade"
    headers["Sec-WebSocket-Accept"] = accept_key(key)



import logging
from types import FrameType, ModuleType
from typing import Literal, Callable, Any, Protocol, Optional, NamedTuple, TypeVar


class Pair(NamedTuple):
    first: Any
    second: Any


T_event = Literal['call', 'line', 'return', 'exception', 'opcode']


T_frame = FrameType


T_tracefunc = Callable[[FrameType, str, Any], Callable[[FrameType, str, Any], Any] | None]


T_tracer_callback_func = Callable[[T_frame, T_event, Any], str]


class TPlugin(Protocol):

    def __init__(self, *args, **kwargs):
        pass

    def on_preprocess(self, *args, **kwargs):
        pass

    def on_postprocess(self, *args, **kwargs):
        pass

    def tracer_callback(self, frame: T_frame, event: T_event, args: Any) -> Any:
        pass


LoggerLike = TypeVar('LoggerLike', bound=logging.Logger)



import os
import re
from typing import Any, Callable, List

from diagnose.engine import PluginManager
from diagnose.typings import TPlugin, T_frame, T_event


@PluginManager.add_plugin(enabled=True)
class ScopePlugin(TPlugin):

    def set_scope_funcs(self, scope_funcs: List[Callable[[str], bool]]):
        assert isinstance(scope_funcs, list)
        self.scope_funcs.extend(scope_funcs)

    @staticmethod
    def _mangle_path(path: str) -> str:
        if re.match(r'^<.*>$', path) is not None:
            return f'inner file {path}'
        return os.path.abspath(path)

    def __init__(self):
        self.scope_funcs = list()

    def on_preprocess(self, *args, **kwargs):
        pass

    def on_postprocess(self, *args, **kwargs):
        pass

    def tracer_callback(self, frame: T_frame, event: T_event, args: Any) -> bool:
        for func in self.scope_funcs:
            if not func(self._mangle_path(frame.f_code.co_filename)):
                return False
        return True



import json

from diagnose.queues import TraceMessageEntry


def parse_from_trace(trace_message_entry: TraceMessageEntry) -> str:
    return json.dumps({
        'id': trace_message_entry.id,
        'filename': trace_message_entry.filename,
        'lineno': trace_message_entry.lineno,
        'classname': trace_message_entry.classsname,
        'file_classname': trace_message_entry.file_classname,
        'cb_rts': trace_message_entry.cb_rts
    })




import asyncio
from typing import Iterable, Mapping, MutableMapping, NewType, Tuple, Union


ConnectionOption = NewType("ConnectionOption", str)


UpgradeProtocol = NewType("UpgradeProtocol", str)


HeadersLike = Union[
    MutableMapping[str, str],
    Mapping[str, str],
    Iterable[Tuple[str, str]]
]


Data = Union[str, bytes]


ServerLike = NewType('ServerLike', asyncio.Server)


BytesLike = bytes, bytearray, memoryview


# Copied from https://github.com/nedbat/coveragepy/
import re
from typing import Dict, Any, Sequence, Mapping

__all__ = ['Template']


class CodeBuilder:
    """
    Build source code conveniently.
    
    The class is for build inner Python code in template. For more details
    see this blog_.
    
    .. _blog: https://aosabook.org/en/500L/a-template-engine.html

    Attributes:
        code (List): The code fragments stored in the list.
        indent_level (int): The current indent level.
            Note that the :attr:`indent_level` must be 0 if the template
            need to render.

    """

    INDENT_STEP = 4

    def __init__(self, indent: int = 0):
        self.code = []
        self.indent_level = indent
    
    def add_line(self, line: str) -> None:
        """
        Add a line of source to the code.

        Indentation and newline will be added for you, don't provide them.

        Args:
            line: A line of Python code

        """
        self.code.extend([" " * self.indent_level, line, "\n"])

    def indent(self):
        """Increase the current indent for following lines."""
        self.indent_level += self.INDENT_STEP

    def dedent(self):
        """Decrease the current indent for following lines."""
        self.indent_level -= self.INDENT_STEP
    
    def add_section(self) -> 'CodeBuilder':
        """
        Add a section, a sub-CodeBuilder.
        
        Returns:
            CodeBuilder: A :class:`CodeBuilder` for injecting Python code.

        """
        section = CodeBuilder(self.indent_level)
        self.code.append(section)
        return section
    
    def __str__(self):
        return "".join(str(c) for c in self.code)
    
    def get_globals(self) -> Dict[str, Any]:
        """
        Execute the code, and return a dict of globals it defines.

        Returns:
            Dict: The ``globals`` of the executed Python code.

        """
        # A check that the caller really finished all the blocks they started.
        assert self.indent_level == 0
        # Get the Python source as a single string.
        python_source = str(self)
        # Execute the source, defining globals, and return them.
        global_namespace = {}
        exec(python_source, global_namespace)
        return global_namespace


class Template:
    """
    A :class:`Template` class for parsing and rendering HTML templates

    Attributes:
        context: The context variables for rendering templates.
        all_vars: All variables used in the template.
        loop_vars: Variables declared in the loop, which means they will
                not be initialized by the context.
    """

    def __init__(self, text, *contexts):
        """Construct a Template with the given `text`.

        Args:
            text: The template text.
            contexts: Dictionaries of values to use for future renderings.
                These are good for filters and global values.

        """
        self.context = {}
        for context in contexts:
            self.context.update(context)
        self.all_vars = set()
        self.loop_vars = set()

        code = CodeBuilder()

        code.add_line("def render_function(context, do_dots):")
        code.indent()
        vars_code = code.add_section()
        code.add_line("result = []")
        code.add_line("append_result = result.append")
        code.add_line("extend_result = result.extend")
        code.add_line("to_str = str")

        buffered = []
        def flush_output():
            """Force `buffered` to the code builder."""
            if len(buffered) == 1:
                code.add_line("append_result(%s)" % buffered[0])
            elif len(buffered) > 1:
                code.add_line("extend_result([%s])" % ", ".join(buffered))
            del buffered[:]
        
        ops_stack = []
        tokens = re.split(r"(?s)({{.*?}}|{%.*?%}|{#.*?#})", text)

        for token in tokens:
            if token.startswith('{#'):
                # Comment: ignore it and move on.
                continue
            elif token.startswith('{{'):
                # An expression to evaluate.
                expr = self._expr_code(token[2:-2].strip())
                buffered.append("to_str(%s)" % expr)
            elif token.startswith('{%'):
                # Action tag: split into words and parse further.
                flush_output()
                words = token[2:-2].strip().split()
                if words[0] == 'if':
                    # An if statement: evaluate the expression to determine if.
                    if len(words) != 2:
                        self._syntax_error("Don't understand if", token)
                    ops_stack.append('if')
                    code.add_line("if %s:" % self._expr_code(words[1]))
                    code.indent()
                elif words[0] == 'elif':
                    if ops_stack[-1] != 'if':
                        self._syntax_error("Missmatched elif", token)
                    else:
                        code.dedent()
                    if len(words) != 2:
                        self._syntax_error("Don't understand elif", token)
                    code.add_line("elif %s:" % self._expr_code(words[1]))
                    code.indent()
                elif words[0] == 'else':
                    if ops_stack[-1] != 'if':
                        self._syntax_error("Missmatched else", token)
                    else:
                        code.dedent()
                    code.add_line("else:")
                    code.indent()
                elif words[0] == 'for':
                    # A loop: iterate over expression result.
                    if len(words) != 4 or words[2] != 'in':
                        self._syntax_error("Don't understand for", token)
                    ops_stack.append('for')
                    self._variable(words[1], self.loop_vars)
                    code.add_line(
                        "for c_%s in %s:" % (
                            words[1],
                            self._expr_code(words[3])
                        )
                    )
                    code.indent()
                elif words[0].startswith('end'):
                    # Endsomething.  Pop the ops stack.
                    if len(words) != 1:
                        self._syntax_error("Don't understand end", token)
                    end_what = words[0][3:]
                    if not ops_stack:
                        self._syntax_error("Too many ends", token)
                    start_what = ops_stack.pop()
                    if start_what != end_what:
                        self._syntax_error("Mismatched end tag", end_what)
                    code.dedent()
                else:
                    self._syntax_error("Don't understand tag", words[0])
            else:
                # Literal content.  If it isn't empty, output it.
                if token:
                    buffered.append(repr(token))

        if ops_stack:
            self._syntax_error("Unmatched action tag", ops_stack[-1])

        flush_output()
        for var_name in self.all_vars - self.loop_vars:
            vars_code.add_line("c_%s = context[%r]" % (var_name, var_name))
        code.add_line("return ''.join(result)")
        code.dedent()
        self._render_function = code.get_globals()['render_function']

    def _syntax_error(self, msg, thing):
        """
        Raise a syntax error using `msg`, and showing `thing`.
        
        """
        raise RuntimeError("%s: %r" % (msg, thing))

    def _expr_code(self, expr):
        """
        Generate a Python expression for `expr`.
        
        Args:
            expr: The expression for executing.
        
        """
        if "|" in expr:
            pipes = expr.split("|")
            code = self._expr_code(pipes[0])
            for func in pipes[1:]:
                self._variable(func, self.all_vars)
                code = "c_%s(%s)" % (func, code)
        elif expr.startswith("~"):
            expr = expr.lstrip("~")
            code = "not %s" % self._expr_code(expr)
        elif "." in expr:
            dots = expr.split(".")
            code = self._expr_code(dots[0])
            args = ", ".join(repr(d) for d in dots[1:])
            code = "do_dots(%s, %s)" % (code, args)
        else:
            self._variable(expr, self.all_vars)
            code = "c_%s" % expr
        return code
    
    def _variable(self, name, vars_set):
        """
        Track that `name` is used as a variable.
        
        Adds the name to `vars_set`, a set of variable names.

        Args:
            name: The variable name.
            vars_set: The variable set.

        Raises:
            RuntimeError: An syntax error if `name` is not a valid name.

        """
        if not re.match(r"[_a-zA-Z][_a-zA-Z0-9]*$", name):
            self._syntax_error("Not a valid name", name)
        vars_set.add(name)
    
    def render(self, context: Dict[str, str] = None) -> str:
        """
        Render this template by applying it to `context`.

        Args:
            context: A dictionary of values to use in this rendering.
        
        Returns:
            str: The rendered result of the template

        """
        # Make the complete context we'll use.
        render_context = dict(self.context)
        if context:
            render_context.update(context)
        return self._render_function(render_context, self._do_dots)
    
    def _do_dots(self, value, *dots):
        """
        Evaluate dotted expressions at runtime.
        
        """
        for dot in dots:
            try:
                value = getattr(value, dot)
            except AttributeError:
                if isinstance(value, Sequence):
                    value = value[int(dot)]
                elif isinstance(value, Mapping):
                    value = value[str(dot)]
            if callable(value):
                value = value()
        return value


"""
Copyright (c) Aymeric Augustin and contributors

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import dataclasses
import enum
import io
import secrets
import struct
import sys
from typing import Callable, Optional, Tuple, Any, Awaitable

from diagnose.server.ws.exception import WebsocketException
from diagnose.server.ws.typings import BytesLike, Data


def apply_mask(data: bytes, mask: bytes) -> bytes:
    """
    Apply masking to the data of a WebSocket message.

    Args:
        data: data to mask.
        mask: 4-bytes mask.

    """
    if len(mask) != 4:
        raise ValueError("mask must contain 4 bytes")

    data_int = int.from_bytes(data, sys.byteorder)
    mask_repeated = mask * (len(data) // 4) + mask[: len(data) % 4]
    mask_int = int.from_bytes(mask_repeated, sys.byteorder)
    return (data_int ^ mask_int).to_bytes(len(data), sys.byteorder)


class Opcode(enum.IntEnum):
    """Opcode values for WebSocket frames."""

    CONT, TEXT, BINARY = 0x00, 0x01, 0x02
    CLOSE, PING, PONG = 0x08, 0x09, 0x0A


OP_CONT = Opcode.CONT
OP_TEXT = Opcode.TEXT
OP_BINARY = Opcode.BINARY
OP_CLOSE = Opcode.CLOSE
OP_PING = Opcode.PING
OP_PONG = Opcode.PONG

DATA_OPCODES = OP_CONT, OP_TEXT, OP_BINARY
CTRL_OPCODES = OP_CLOSE, OP_PING, OP_PONG


class CloseCode(enum.IntEnum):
    """Close code values for WebSocket close frames."""

    NORMAL_CLOSURE = 1000
    GOING_AWAY = 1001
    PROTOCOL_ERROR = 1002
    UNSUPPORTED_DATA = 1003
    # 1004 is reserved
    NO_STATUS_RCVD = 1005
    ABNORMAL_CLOSURE = 1006
    INVALID_DATA = 1007
    POLICY_VIOLATION = 1008
    MESSAGE_TOO_BIG = 1009
    MANDATORY_EXTENSION = 1010
    INTERNAL_ERROR = 1011
    SERVICE_RESTART = 1012
    TRY_AGAIN_LATER = 1013
    BAD_GATEWAY = 1014
    TLS_HANDSHAKE = 1015


# See https://www.iana.org/assignments/websocket/websocket.xhtml
CLOSE_CODE_EXPLANATIONS: dict[int, str] = {
    CloseCode.NORMAL_CLOSURE: "OK",
    CloseCode.GOING_AWAY: "going away",
    CloseCode.PROTOCOL_ERROR: "protocol error",
    CloseCode.UNSUPPORTED_DATA: "unsupported data",
    CloseCode.NO_STATUS_RCVD: "no status received [internal]",
    CloseCode.ABNORMAL_CLOSURE: "abnormal closure [internal]",
    CloseCode.INVALID_DATA: "invalid frame payload data",
    CloseCode.POLICY_VIOLATION: "policy violation",
    CloseCode.MESSAGE_TOO_BIG: "message too big",
    CloseCode.MANDATORY_EXTENSION: "mandatory extension",
    CloseCode.INTERNAL_ERROR: "internal error",
    CloseCode.SERVICE_RESTART: "service restart",
    CloseCode.TRY_AGAIN_LATER: "try again later",
    CloseCode.BAD_GATEWAY: "bad gateway",
    CloseCode.TLS_HANDSHAKE: "TLS handshake failure [internal]",
}


# Close code that are allowed in a close frame.
# Using a set optimizes `code in EXTERNAL_CLOSE_CODES`.
EXTERNAL_CLOSE_CODES = {
    CloseCode.NORMAL_CLOSURE,
    CloseCode.GOING_AWAY,
    CloseCode.PROTOCOL_ERROR,
    CloseCode.UNSUPPORTED_DATA,
    CloseCode.INVALID_DATA,
    CloseCode.POLICY_VIOLATION,
    CloseCode.MESSAGE_TOO_BIG,
    CloseCode.MANDATORY_EXTENSION,
    CloseCode.INTERNAL_ERROR,
    CloseCode.SERVICE_RESTART,
    CloseCode.TRY_AGAIN_LATER,
    CloseCode.BAD_GATEWAY,
}


OK_CLOSE_CODES = {
    CloseCode.NORMAL_CLOSURE,
    CloseCode.GOING_AWAY,
    CloseCode.NO_STATUS_RCVD,
}


@dataclasses.dataclass
class Frame:
    """
    WebSocket frame.

    Attributes:
        opcode: Opcode.
        data: Payload data.
        fin: FIN bit.
        rsv1: RSV1 bit.
        rsv2: RSV2 bit.
        rsv3: RSV3 bit.

    Only these fields are needed. The MASK bit, payload length and masking-key
    are handled on the fly when parsing and serializing frames.

    """

    opcode: Opcode
    data: bytes
    fin: bool = True
    rsv1: bool = False
    rsv2: bool = False
    rsv3: bool = False

    def __str__(self) -> str:
        """
        Return a human-readable representation of a frame.

        """
        coding = None
        length = f"{len(self.data)} byte{'' if len(self.data) == 1 else 's'}"
        non_final = "" if self.fin else "continued"

        if self.opcode is OP_TEXT:
            # Decoding only the beginning and the end is needlessly hard.
            # Decode the entire payload then elide later if necessary.
            data = repr(self.data.decode())
        elif self.opcode is OP_BINARY:
            # We'll show at most the first 16 bytes and the last 8 bytes.
            # Encode just what we need, plus two dummy bytes to elide later.
            binary = self.data
            if len(binary) > 25:
                binary = b"".join([binary[:16], b"\x00\x00", binary[-8:]])
            data = " ".join(f"{byte:02x}" for byte in binary)
        elif self.opcode is OP_CLOSE:
            data = str(Close.parse(self.data))
        elif self.data:
            # We don't know if a Continuation frame contains text or binary.
            # Ping and Pong frames could contain UTF-8.
            # Attempt to decode as UTF-8 and display it as text; fallback to
            # binary. If self.data is a memoryview, it has no decode() method,
            # which raises AttributeError.
            try:
                data = repr(self.data.decode())
                coding = "text"
            except (UnicodeDecodeError, AttributeError):
                binary = self.data
                if len(binary) > 25:
                    binary = b"".join([binary[:16], b"\x00\x00", binary[-8:]])
                data = " ".join(f"{byte:02x}" for byte in binary)
                coding = "binary"
        else:
            data = "''"

        if len(data) > 75:
            data = data[:48] + "..." + data[-24:]

        metadata = ", ".join(filter(None, [coding, length, non_final]))

        return f"{self.opcode.name} {data} [{metadata}]"

    def serialize(
        self,
        *,
        mask: bool
    ) -> bytes:
        """
        Serialize a WebSocket frame.

        Args:
            mask: whether the frame should be masked i.e. whether the write
                happens on the client side.

        Raises:
            WebsocketException: if the frame contains incorrect values.

        """
        self.check()

        output = io.BytesIO()

        # Prepare the header.
        head1 = (
            (0b10000000 if self.fin else 0)
            | (0b01000000 if self.rsv1 else 0)
            | (0b00100000 if self.rsv2 else 0)
            | (0b00010000 if self.rsv3 else 0)
            | self.opcode
        )

        head2 = 0b10000000 if mask else 0

        length = len(self.data)
        if length < 126:
            output.write(struct.pack("!BB", head1, head2 | length))
        elif length < 65536:
            output.write(struct.pack("!BBH", head1, head2 | 126, length))
        else:
            output.write(struct.pack("!BBQ", head1, head2 | 127, length))

        if mask:
            mask_bytes = secrets.token_bytes(4)
            output.write(mask_bytes)

        # Prepare the data.
        if mask:
            data = apply_mask(self.data, mask_bytes)
        else:
            data = self.data
        output.write(data)

        return output.getvalue()

    def check(self) -> None:
        """
        Check that reserved bits and opcode have acceptable values.

        Raises:
            WebsocketException: if a reserved bit or the opcode is invalid.

        """
        if self.rsv1 or self.rsv2 or self.rsv3:
            raise WebsocketException("ProtocolError: reserved bits must be 0")

        if self.opcode in CTRL_OPCODES:
            if len(self.data) > 125:
                raise WebsocketException("ProtocolError: control frame too long")
            if not self.fin:
                raise WebsocketException("ProtocolError: fragmented control frame")

    @classmethod
    async def read(
            cls,
            reader: Callable[[int], Awaitable[bytes]],
            *,
            mask: bool,
            max_size: Optional[int] = None
    ) -> 'Frame':
        """
        Read a WebSocket frame.

        Args:
            reader: coroutine that reads exactly the requested number of
                bytes, unless the end of file is reached.
            mask: whether the frame should be masked i.e. whether the read
                happens on the server side.
            max_size: maximum payload size in bytes.

        Raises:
            WebsocketException: if the frame exceeds ``max_size``.
            WebsocketException: if the frame contains incorrect values.

        """

        # Read the header.
        data = await reader(2)
        head1, head2 = struct.unpack("!BB", data)

        # While not Pythonic, this is marginally faster than calling bool().
        fin = True if head1 & 0b10000000 else False
        rsv1 = True if head1 & 0b01000000 else False
        rsv2 = True if head1 & 0b00100000 else False
        rsv3 = True if head1 & 0b00010000 else False

        try:
            opcode = Opcode(head1 & 0b00001111)
        except ValueError as exc:
            raise WebsocketException("ProtocolError: invalid opcode") from exc

        if (True if head2 & 0b10000000 else False) != mask:
            raise WebsocketException("ProtocolError: incorrect masking")

        length = head2 & 0b01111111
        if length == 126:
            data = await reader(2)
            (length,) = struct.unpack("!H", data)
        elif length == 127:
            data = await reader(8)
            (length,) = struct.unpack("!Q", data)
        if max_size is not None and length > max_size:
            raise WebsocketException(f"PayloadTooBig: over size limit ({length} > {max_size} bytes)")
        if mask:
            mask_bits = await reader(4)

        # Read the data.
        data = await reader(length)
        if mask:
            data = apply_mask(data, mask_bits)

        frame = Frame(opcode, data, fin, rsv1, rsv2, rsv3)
        frame.check()

        return frame

    def write(
            self,
            write: Callable[[bytes], Any],
            *,
            mask: bool
    ) -> None:
        """
        Write a WebSocket frame.

        Args:
            frame: frame to write.
            write: function that writes bytes.
            mask: whether the frame should be masked i.e. whether to write
                happens on the client side.

        Raises:
            WebsocketException: if the frame contains incorrect values.

        """
        # The frame is written in a single call to write in order to prevent
        # TCP fragmentation. See #68 for details. This also makes it safe to
        # send frames concurrently from multiple coroutines.
        write(self.serialize(mask=mask))


def prepare_data(data: Data) -> Tuple[int, bytes]:
    """
    Convert a string or byte-like object to an opcode and a bytes-like object.

    This function is designed for data frames.

    If ``data`` is a :class:`str`, return ``OP_TEXT`` and a :class:`bytes`
    object encoding ``data`` in UTF-8.

    If ``data`` is a bytes-like object, return ``OP_BINARY`` and a bytes-like
    object.

    Raises:
        TypeError: if ``data`` doesn't have a supported type.

    """
    if isinstance(data, str):
        return OP_TEXT, data.encode("utf-8")
    elif isinstance(data, BytesLike):
        return OP_BINARY, data
    else:
        raise TypeError("data must be str or bytes-like")


def prepare_ctrl(data: Data) -> bytes:
    """
    Convert a string or byte-like object to bytes.

    This function is designed for ping and pong frames.

    If ``data`` is a :class:`str`, return a :class:`bytes` object encoding
    ``data`` in UTF-8.

    If ``data`` is a bytes-like object, return a :class:`bytes` object.

    Raises:
        TypeError: if ``data`` doesn't have a supported type.

    """
    if isinstance(data, str):
        return data.encode("utf-8")
    elif isinstance(data, BytesLike):
        return bytes(data)
    else:
        raise TypeError("data must be str or bytes-like")


@dataclasses.dataclass
class Close:
    """
    Code and reason for WebSocket close frames.

    Attributes:
        code: Close code.
        reason: Close reason.

    """

    code: int
    reason: str

    def __str__(self) -> str:
        """
        Return a human-readable representation of a close code and reason.

        """
        if 3000 <= self.code < 4000:
            explanation = "registered"
        elif 4000 <= self.code < 5000:
            explanation = "private use"
        else:
            explanation = CLOSE_CODE_EXPLANATIONS.get(self.code, "unknown")
        result = f"{self.code} ({explanation})"

        if self.reason:
            result = f"{result} {self.reason}"

        return result

    @classmethod
    def parse(cls, data: bytes) -> 'Close':
        """
        Parse the payload of a close frame.

        Args:
            data: payload of the close frame.

        Raises:
            WebsocketException: if data is ill-formed.
            WebsocketException: if the reason isn't valid UTF-8.

        """
        if len(data) >= 2:
            (code,) = struct.unpack("!H", data[:2])
            reason = data[2:].decode("utf-8")
            close = cls(code, reason)
            close.check()
            return close
        elif len(data) == 0:
            return cls(CloseCode.NO_STATUS_RCVD, "")
        else:
            raise WebsocketException("ProtocolError: close frame too short")

    def serialize(self) -> bytes:
        """
        Serialize the payload of a close frame.

        """
        self.check()
        return struct.pack("!H", self.code) + self.reason.encode("utf-8")

    def check(self) -> None:
        """
        Check that the close code has a valid value for a close frame.

        Raises:
            WebsocketException: if the close code is invalid.

        """
        if not (self.code in EXTERNAL_CLOSE_CODES or 3000 <= self.code < 5000):
            raise WebsocketException("ProtocolError: invalid status code")


from diagnose.queues import Q
from diagnose.server.parse import parse_from_trace
from diagnose.server.ws import Websocket


class WebsocketHandler:

    # todo: Use signal to make bi-comunication
    def __init__(self, signal: 'Signal' = None, queue: 'Q' = None):
        self.signal = signal
        self.queue = queue

    async def _do_start(self):
        for _message in Q:
            await self.ws.send(parse_from_trace(_message))

    async def _do_pause(self):
        pass

    async def _do_resume(self):
        pass

    async def _do_stop(self):
        pass

    async def __call__(self, ws: Websocket):
        self.ws = ws
        async for message in ws:
            if message == 'start':
                await self._do_start()
            elif message == 'pause':
                await self._do_pause()
            elif message == 'resume':
                await self._do_resume()
            elif message == 'stop':
                await self._do_stop()
            else:
                raise RuntimeError(f"Unexpected message from clinet {message}")






import getopt
import os.path
import sys

from diagnose.engine import Pipeline
from diagnose.engine import PluginManager


class Cmd:

    usage = '''
    Usages: hot-diagnose [OPTIONS] [OPTION_ARGS] [SOURCE_ARGS]
    
    Options:
    
        -s --source: The source file to execute. If the source file needs arguments, they can be append to the tail.
        -o --output: The path to the output file of the executing source file.
        -p --path:   The directory path that all files in the path will be rendered as output. If not provided, the
                     path to the source file and its directory will be set as default. Usually, it will also regard as
                     PYTHONPATH.
        -h --help:   The help message.
    '''

    def parse(self):
        opts, args = getopt.getopt(sys.argv[1:],
                                   's:o:p:h',
                                   ['source=', 'output', 'path', 'help'])

        if any(opt in ['-h', '--help'] for opt, optarg in opts):
            print(self.usage)
            sys.exit()

        source = [optarg for opt, optarg in opts if opt in ['-s', '--source']]
        assert len(source) == 1
        source = source[0]

        PluginManager.load_plugins(['diagnose.plugins'])

        if any(opt in ['-o', '--output'] for opt, optarg in opts):
            output = [optarg for opt, optarg in opts if opt in ['-o', '--output']]
            if len(output) > 1:
                raise RuntimeError("Cannot support multiple redirect output")
            PluginManager.enable_plugin('RedirectPlugin')
            redirect_plugin = PluginManager.get_plugin('RedirectPlugin')
            redirect_plugin.set_filename(output[0])

        if any(opt in ['-p', '--path'] for opt, optarg in opts):
            scope_paths = [optarg for opt, optarg in opts if opt in ['-p', '--path']]
        else:
            scope_paths = [source, os.path.dirname(source)]

        scope_plugin = PluginManager.get_plugin('ScopePlugin')
        _funcs = []
        for path in scope_paths:
            if not os.path.isabs(path):
                path = os.path.abspath(path)
            sys.path.append(path)
            _funcs.append(lambda p: p.startswith(path))
        scope_plugin.set_scope_funcs(_funcs)

        pipeline = Pipeline(source, args, scope_paths[0])
        pipeline.run()


import hashlib
import ntpath
import os.path
import pathlib
import re
import time
from os import PathLike
from typing import Tuple


def read_source_py(filename: str) -> bytes:  # also for pyw
    filename_base, filename_ext = os.path.splitext(filename)
    assert filename_ext == '.py' or filename_ext == '.pyw'
    if not os.path.exists(filename):
        raise IOError(f"Cannot find target filename: {filename}")
    with open(filename, 'rb') as f:
        source = f.read()
    return source.replace(b"\r\n", b"\n").replace(b"\r", b"\n")


def read_source(filename: str) -> bytes:
    if not os.path.exists(filename):
        raise IOError(f"Cannot find target filename: {filename}")
    with open(filename, 'rb') as f:
        source = f.read()
    return source.replace(b"\r\n", b"\n").replace(b"\r", b"\n")


def read_source_py_with_line(filename: str) -> Tuple[int, str]:
    filename_base, filename_ext = os.path.splitext(filename)
    assert filename_ext == '.py' or filename_ext == '.pyw'
    if not os.path.exists(filename):
        raise IOError(f"Cannot find target filename: {filename}")
    lineno = 0
    with open(filename, 'rb') as f:
        while line := f.readline():
            lineno += 1
            yield lineno, line


def get_home_dir() -> str:
    return str(pathlib.Path.home())


def get_package_dir() -> str:
    return __file__.rsplit(os.path.sep, 1)[0]


def mkdir(location: PathLike | str, dirname: str):
    path = os.path.join(location, dirname)
    if os.path.exists(path):
        return path
    os.mkdir(path)
    return path


def write_file(filename: str, content: str | bytes):
    content = re.sub(r"(\A\s+)|(\s+$)", "", content, flags=re.MULTILINE) + "\n"
    with open(filename, "wb") as fout:
        fout.write(content.encode("ascii", "xmlcharrefreplace"))


def generate_classname(full_pathname: str, lineno: int = 0):
    assert os.path.isabs(full_pathname) and isinstance(lineno, int)
    if os.path.isdir(full_pathname):
        return hashlib.new('sha3_256', full_pathname.encode('UTF-8')).hexdigest()[:16]
    elif os.path.isfile(full_pathname):
        return hashlib.new('sha3_256', f'{full_pathname}:{lineno}'.encode('UTF-8')).hexdigest()[:16]


def stat(path: str) -> Tuple[str, ...]:
    assert os.path.abspath(path)
    st = os.stat(path)
    size_bytes = st.st_size
    units = ['B', 'KB', 'MB', 'GB']
    u_index = 0
    while size_bytes >= 1024:
        size_bytes /= 1024
        u_index += 1
    size = str(size_bytes) + units[u_index]
    mtime_sec = st.st_mtime
    mtime = time.strftime("%Y-%m-%d", time.localtime(mtime_sec))
    return mtime, size


import asyncio
import functools
import socket
import http.client  # must be imported
from types import TracebackType
from typing import Optional, Set, Iterable, Type, Union, Callable, Awaitable, Any, Sequence, Generator

from diagnose.logs import Logger
from diagnose.server.ws.misc import State
from diagnose.server.ws.protocol import WebSocketServerProtocol
from diagnose.typings import LoggerLike


class WebSocketServer:
    """
    WebSocket server returned by :func:`serve`.

    This class provides the same interface as :class:`~asyncio.Server`,
    notably the :meth:`~asyncio.Server.close`
    and :meth:`~asyncio.Server.wait_closed` methods.

    It keeps track of WebSocket connections in order to close them properly
    when shutting down.

    Args:
        logger: logger for this server;
            defaults to ``logging.getLogger("websockets.server")``.

    """

    def __init__(self, logger: Optional['LoggerLike'] = None):
        if logger is None:
            logger = Logger.get_logger("websockets.server")
        self.logger = logger

        # Keep track of active connections.
        self.websockets: Set['WebSocketServerProtocol'] = set()

        # Task responsible for closing the server and terminating connections.
        self.close_task: Optional[asyncio.Task[None]] = None

        # Completed when the server is closed and connections are terminated.
        self.closed_waiter: asyncio.Future[None]

    def wrap(self, server: asyncio.base_events.Server) -> None:
        """
        Attach to a given :class:`~asyncio.Server`.

        Since :meth:`~asyncio.loop.create_server` doesn't support injecting a
        custom ``Server`` class, the easiest solution that doesn't rely on
        private :mod:`asyncio` APIs is to:

        - instantiate a :class:`WebSocketServer`
        - give the protocol factory a reference to that instance
        - call :meth:`~asyncio.loop.create_server` with the factory
        - attach the resulting :class:`~asyncio.Server` with this method

        """
        self.server = server
        for sock in server.sockets:
            if sock.family == socket.AF_INET:
                name = "%s:%d" % sock.getsockname()
            elif sock.family == socket.AF_INET6:
                name = "[%s]:%d" % sock.getsockname()[:2]
            elif sock.family == socket.AF_UNIX:
                name = sock.getsockname()
            # In the unlikely event that someone runs websockets over a
            # protocol other than IP or Unix sockets, avoid crashing.
            else:
                name = str(sock.getsockname())
            self.logger.info("server listening on %s", name)

        self.closed_waiter = server.get_loop().create_future()

    def register(self, protocol: 'WebSocketServerProtocol') -> None:
        """
        Register a connection with this server.

        """
        self.websockets.add(protocol)

    def unregister(self, protocol: 'WebSocketServerProtocol') -> None:
        """
        Unregister a connection with this server.

        """
        self.websockets.remove(protocol)

    def close(self, close_connections: bool = True) -> None:
        """
        Close the server.

        * Close the underlying :class:`~asyncio.Server`.
        * When ``close_connections`` is :obj:`True`, which is the default,
          close existing connections. Specifically:

          * Reject opening WebSocket connections with an HTTP 503 (service
            unavailable) error. This happens when the server accepted the TCP
            connection but didn't complete the opening handshake before closing.
          * Close open WebSocket connections with close code 1001 (going away).

        * Wait until all connection handlers terminate.

        :meth:`close` is idempotent.

        """
        if self.close_task is None:
            self.close_task = self.get_loop().create_task(
                self._close(close_connections)
            )

    async def _close(self, close_connections: bool) -> None:
        """
        Implementation of :meth:`close`.

        This calls :meth:`~asyncio.Server.close` on the underlying
        :class:`~asyncio.Server` object to stop accepting new connections and
        then closes open connections with close code 1001.

        """
        self.logger.info("server closing")

        # Stop accepting new connections.
        self.server.close()

        # Wait until self.server.close() completes.
        await self.server.wait_closed()

        # Wait until all accepted connections reach connection_made() and call
        # register(). See https://bugs.python.org/issue34852 for details.
        await asyncio.sleep(0)

        if close_connections:
            # Close OPEN connections with status code 1001. Since the server was
            # closed, handshake() closes OPENING connections with an HTTP 503
            # error. Wait until all connections are closed.

            close_tasks = [
                asyncio.create_task(websocket.close(1001))
                for websocket in self.websockets
                if websocket.state is not State.CONNECTING
            ]
            # asyncio.wait doesn't accept an empty first argument.
            if close_tasks:
                await asyncio.wait(
                    close_tasks
                )

        # Wait until all connection handlers are complete.

        # asyncio.wait doesn't accept an empty first argument.
        if self.websockets:
            await asyncio.wait(
                [websocket.handler_task for websocket in self.websockets]
            )

        # Tell wait_closed() to return.
        self.closed_waiter.set_result(None)

        self.logger.info("server closed")

    async def wait_closed(self) -> None:
        """
        Wait until the server is closed.

        When :meth:`wait_closed` returns, all TCP connections are closed and
        all connection handlers have returned.

        """
        await asyncio.shield(self.closed_waiter)

    def get_loop(self) -> asyncio.AbstractEventLoop:
        """
        See :meth:`asyncio.Server.get_loop`.

        """
        return self.server.get_loop()

    def is_serving(self) -> bool:
        """
        See :meth:`asyncio.Server.is_serving`.

        """
        return self.server.is_serving()

    async def start_serving(self) -> None:
        """
        See :meth:`asyncio.Server.start_serving`.

        Typical use::

            server = await serve(..., start_serving=False)
            # perform additional setup here...
            # ... then start the server
            await server.start_serving()

        """
        await self.server.start_serving()  # pragma: no cover

    async def serve_forever(self) -> None:
        """
        See :meth:`asyncio.Server.serve_forever`.

        Typical use::

            server = await serve(...)
            # this coroutine doesn't return
            # canceling it stops the server
            await server.serve_forever()

        This is an alternative to using :func:`serve` as an asynchronous context
        manager. Shutdown is triggered by canceling :meth:`serve_forever`
        instead of exiting a :func:`serve` context.

        """
        await self.server.serve_forever()  # pragma: no cover

    @property
    def sockets(self) -> Iterable[socket.socket]:
        """
        See :attr:`asyncio.Server.sockets`.

        """
        return self.server.sockets

    async def __aenter__(self) -> 'WebSocketServer':
        return self  # pragma: no cover

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> None:
        self.close()
        await self.wait_closed()


class Serve:
    """
    Start a WebSocket server listening on ``host`` and ``port``.

    Whenever a client connects, the server creates a
    :class:`WebSocketServerProtocol`, performs the opening handshake, and
    delegates to the connection handler, ``ws_handler``.

    The handler receives the :class:`WebSocketServerProtocol` and uses it to
    send and receive messages.

    Once the handler completes, either normally or with an exception, the
    server performs the closing handshake and closes the connection.

    Awaiting :func:`serve` yields a :class:`WebSocketServer`. This object
    provides a :meth:`~WebSocketServer.close` method to shut down the server::

        stop = asyncio.Future()  # set this future to exit the server

        server = await serve(...)
        await stop
        await server.close()

    :func:`serve` can be used as an asynchronous context manager. Then, the
    server is shut down automatically when exiting the context::

        stop = asyncio.Future()  # set this future to exit the server

        async with serve(...):
            await stop

    Args:
        ws_handler: connection handler. It receives the WebSocket connection,
            which is a :class:`WebSocketServerProtocol`, in argument.
        host: network interfaces the server is bound to;
            see :meth:`~asyncio.loop.create_server` for details.
        port: TCP port the server listens on;
            see :meth:`~asyncio.loop.create_server` for details.
        logger: logger for this server;
            defaults to ``logging.getLogger("websockets.server")``.
        
    Returns:
        WebSocketServer: WebSocket server.

    """

    def __init__(
            self,
            ws_handler: Callable[['WebSocketServerProtocol'], Awaitable[Any]],
            host: Optional[Union[str, Sequence[str]]] = None,
            port: Optional[int] = None,
            *,
            logger: Optional['LoggerLike'] = Logger.get_logger("websockets.server"),
            ping_interval: Optional[float] = 20,
            ping_timeout: Optional[float] = 20,
            max_size: Optional[int] = 2 ** 20,
            max_queue: Optional[int] = 2 ** 5,
            read_limit: int = 2 ** 16,
            write_limit: int = 2 ** 16,
            **kwargs: Any,
    ) -> None:
        close_timeout = 10

        ws_server = WebSocketServer(logger=logger)

        loop = asyncio.get_event_loop()

        self.loop = loop

        factory = functools.partial(
            WebSocketServerProtocol,
            ws_handler,
            ws_server,
            host=host,
            port=port,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            close_timeout=close_timeout,
            max_size=max_size,
            max_queue=max_queue,
            read_limit=read_limit,
            write_limit=write_limit,
            loop=loop,
            legacy_recv=False,
            logger=logger,
        )

        if kwargs.pop("unix", False):
            path: Optional[str] = kwargs.pop("path", None)
            # unix_serve(path) must not specify host and port parameters.
            assert host is None and port is None
            create_server = functools.partial(
                loop.create_unix_server, factory, path, **kwargs
            )
        else:
            create_server = functools.partial(
                loop.create_server, factory, host, port, **kwargs
            )

        # This is a coroutine function.
        self._create_server = create_server
        self.ws_server = ws_server

    async def __aenter__(self) -> WebSocketServer:
        return await self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> None:
        self.ws_server.close()
        await self.ws_server.wait_closed()

    def __await__(self) -> Generator[Any, None, WebSocketServer]:
        # Create a suitable iterator by calling __await__ on a coroutine.
        return self.__await_impl__().__await__()

    async def __await_impl__(self) -> WebSocketServer:
        server = await self._create_server()
        self.ws_server.wrap(server)
        return self.ws_server

    __iter__ = __await__


serve = Serve



"""
This package implements the smallest websocket according to `RFC 6455`_.

The code in this package is mostly from the famous Python project `websockets`_
by Aymeric Augustin and other contributors. To make the websocket simple and
fit the senario of the current project, a lot of details and features are removed, 
including origin checking, extensions, extra headers, and request processing, etc.
Actually, it only covers the server-side websocket implementation and all exceptions
appearing in the project are flattened into :exc:`WebsocketException`.

As described in the ``websockets``, the using of the websocket framwork is quite
easy::

    from server.ws import Websocket, serve


    async def echo(ws: Websocket):
        async for message in ws:
            await ws.send(message)

    async def main():
        async with serve(echo, "localhost", 8765):
            await asyncio.Future()

    asyncio.run(main())


.. _`RFC 6455`: https://datatracker.ietf.org/doc/html/rfc6455.html
.. _`websockets`: https://github.com/python-websockets/websockets
"""

from diagnose.server.ws.protocol import WebSocketServerProtocol as Websocket
from diagnose.server.ws.serve import serve
from diagnose.server.ws.exception import WebsocketException


__all__ = ['Websocket', 'serve', 'WebsocketException']


import importlib
from dataclasses import dataclass, field
from typing import Optional, ClassVar, Dict, Type, Sequence

from diagnose.logs import Logger
from diagnose.typings import TPlugin


@dataclass
class PluginDescriptor:
    enable: bool = field(default=False)
    plugin: TPlugin = field(default=None)

    def __repr__(self):
        return f'<PluginDescriptor enable={self.enable}> plugin={self.plugin.__name__}>'


class PluginManager:
    plugins: ClassVar[Dict[str, PluginDescriptor]] = dict()

    @classmethod
    def add_plugin(cls, enabled: bool = False, *args, **kwargs):
        def _inner(plugin_cls: Type['TPlugin']):
            if plugin_cls.__name__ in cls.plugins.keys():
                raise RuntimeError(f"The plugin {plugin_cls.__name__} is duplicated")
            plugin = plugin_cls(*args, **kwargs)
            cls.plugins[plugin_cls.__name__] = PluginDescriptor(enable=enabled, plugin=plugin)

        return _inner

    @classmethod
    def load_plugins(cls, plugin_names: Sequence[str]):
        for plugin_name in plugin_names:
            try:
                importlib.import_module(plugin_name)
            except ModuleNotFoundError:
                Logger.get_logger('engine').error(f"Cannot find plugin {plugin_name}", exc_info=True)

    @classmethod
    def enable_plugin(cls, plugin_cls_name: str) -> None:
        for plugin_name, plugin_disc in cls.plugins.items():
            if plugin_name == plugin_cls_name:
                cls.plugins[plugin_name].enable = True
                return
        Logger.get_logger('engine').warning(f"Cannot find target plugin {plugin_cls_name}")

    @classmethod
    def get_plugin(cls, plugin_cls_name: str) -> Optional['TPlugin']:
        for plugin_name, plugin_disc in cls.plugins.items():
            if plugin_name == plugin_cls_name:
                return plugin_disc.plugin
        Logger.get_logger('engine').warning(f"Cannot find target plugin {plugin_cls_name}")
        return None



import queue
from dataclasses import dataclass, field

from typing import List, TypeVar, Any, Tuple, ClassVar

__all__ = ['TraceMessageEntry', 'ActionMessageEntry', 'Q']


class MessageEntry:
    type: str

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, item):
        return self.__getattribute__(item)


@dataclass
class TraceMessageEntry(MessageEntry):
    id: int
    filename: str
    lineno: int
    classsname: int
    file_classname: str
    cb_rts: str
    type: str = field(default='TraceMessage')

    def __repr__(self):
        return f"<TraceMessageEntry id={self.id} filename={self.filename} lineno={self.lineno} cb_rts={self.cb_rts}>"


@dataclass
class ActionMessageEntry(MessageEntry):
    id: int
    action: str
    value: str
    type: str = field(default='ActionMessage')

    def __repr__(self):
        return f"<ActionMessageEntry id={self.id} action={self.action} value={self.value}>"


T = TypeVar('T', bound=MessageEntry)


class MessageQueue(queue.Queue):
    _self_id: int = 0
    _get_id: int = -1

    queue: List[T]

    def _init(self, maxsize):
        self.queue = list()

    def _put(self, message_entry: T) -> None:
        message_entry.id = self._self_id
        self._self_id += 1
        self.queue.append(message_entry)

    def _get(self) -> T:
        return self.queue.pop(0)

    def _qsize(self) -> int:
        return len(self.queue)

    def clear(self):
        self._self_id = 0
        self.queue.clear()

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.queue.pop(0)
        except IndexError:
            raise StopIteration
        

Q = MessageQueue()



"""
Copyright (c) Aymeric Augustin and contributors

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import re
from typing import Callable, List, Optional, Tuple, TypeVar, cast

from diagnose.server.ws.typings import ConnectionOption, UpgradeProtocol


def peek_ahead(header: str, pos: int) -> Optional[str]:
    """
    Return the next character from ``header`` at the given position.

    Return :obj:`None` at the end of ``header``.

    We never need to peek more than one character ahead.

    """
    return None if pos == len(header) else header[pos]


_OWS_re = re.compile(r"[\t ]*")


def parse_OWS(header: str, pos: int) -> int:
    """
    Parse optional whitespace from ``header`` at the given position.

    Return the new position.

    The whitespace itself isn't returned because it isn't significant.

    """
    # There's always a match, possibly empty, whose content doesn't matter.
    match = _OWS_re.match(header, pos)
    assert match is not None
    return match.end()


T = TypeVar('T')


def parse_list(
    parse_item: Callable[[str, int, str], Tuple[T, int]],
    header: str,
    pos: int,
    header_name: str,
) -> List[T]:
    """
    Parse a comma-separated list from ``header`` at the given position.

    This is appropriate for parsing values with the following grammar:

        1#item

    ``parse_item`` parses one item.

    ``header`` is assumed not to start or end with whitespace.

    (This function is designed for parsing an entire header value and
    :func:`read_headers` strips whitespace from values.)

    Return a list of items.

    Raises:
        RuntimeError: on invalid inputs.

    """
    # Per https://www.rfc-editor.org/rfc/rfc7230.html#section-7, "a recipient
    # MUST parse and ignore a reasonable number of empty list elements";
    # hence while loops that remove extra delimiters.

    # Remove extra delimiters before the first item.
    while peek_ahead(header, pos) == ",":
        pos = parse_OWS(header, pos + 1)

    items = []
    while True:
        # Loop invariant: an item starts at pos in header.
        item, pos = parse_item(header, pos, header_name)
        items.append(item)
        pos = parse_OWS(header, pos)

        # We may have reached the end of the header.
        if pos == len(header):
            break

        # There must be a delimiter after each element except the last one.
        if peek_ahead(header, pos) == ",":
            pos = parse_OWS(header, pos + 1)
        else:
            raise RuntimeError(
                header_name, "expected comma", header, pos
            )

        # Remove extra delimiters before the next item.
        while peek_ahead(header, pos) == ",":
            pos = parse_OWS(header, pos + 1)

        # We may have reached the end of the header.
        if pos == len(header):
            break

    # Since we only advance in the header by one character with peek_ahead()
    # or with the end position of a regex match, we can't overshoot the end.
    assert pos == len(header)

    return items


_token_re = re.compile(r"[-!#$%&\'*+.^_`|~0-9a-zA-Z]+")


def parse_token(header: str, pos: int, header_name: str) -> Tuple[str, int]:
    """
    Parse a token from ``header`` at the given position.

    Return the token value and the new position.

    Raises:
        RuntimeError: on invalid inputs.

    """
    match = _token_re.match(header, pos)
    if match is None:
        raise RuntimeError(header_name, "expected token", header, pos)
    return match.group(), match.end()


def parse_connection_option(
    header: str, pos: int, header_name: str
) -> Tuple[ConnectionOption, int]:
    """
    Parse a Connection option from ``header`` at the given position.

    Return the protocol value and the new position.

    Raises:
        RuntimeError: on invalid inputs.

    """
    item, pos = parse_token(header, pos, header_name)
    return cast(ConnectionOption, item), pos


def parse_connection(header: str) -> List[ConnectionOption]:
    """
    Parse a ``Connection`` header.

    Return a list of HTTP connection options.

    Args
        header: value of the ``Connection`` header.

    Raises:
        RuntimeError: on invalid inputs.

    """
    return parse_list(parse_connection_option, header, 0, "Connection")


_protocol_re = re.compile(
    r"[-!#$%&\'*+.^_`|~0-9a-zA-Z]+(?:/[-!#$%&\'*+.^_`|~0-9a-zA-Z]+)?"
)


def parse_upgrade_protocol(
    header: str, pos: int, header_name: str
) -> Tuple[UpgradeProtocol, int]:
    """
    Parse an Upgrade protocol from ``header`` at the given position.

    Return the protocol value and the new position.

    Raises:
        RuntimeError: on invalid inputs.

    """
    match = _protocol_re.match(header, pos)
    if match is None:
        raise RuntimeError(
            f"{header_name}, 'expected protocol', {header}, {pos}"
        )
    return cast(UpgradeProtocol, match.group()), match.end()


def parse_upgrade(header: str) -> List[UpgradeProtocol]:
    """
    Parse an ``Upgrade`` header.

    Return a list of HTTP protocols.

    Args:
        header: value of the ``Upgrade`` header.

    Raises:
        RuntimeError: on invalid inputs.

    """
    return parse_list(parse_upgrade_protocol, header, 0, "Upgrade")


import os.path
import sys
from typing import Any, Optional, List
from types import ModuleType

from diagnose.fileutils import read_source_py
from diagnose.engine.tracer import Tracer
from diagnose.typings import T_tracer_callback_func, LoggerLike

__all__ = ['PyRunner']


class DummyLoader:
    """
    The loader for the dummy module
    """

    def __init__(self, fullname: str, *args: Any) -> None:
        self.fullname = fullname


class PyRunner:
    """
    The :class:`PyRunner` is for running Python script or package from the
    cammand line.

    The idea of the class is first compiling the given source files and then
    executing the :obj:`code` with the args from the command line. When
    executing the :obj:`code`, a trace is set to collect the runtime
    information. The collected information  will be sent to the :class:`Q`
    for further using.

    Attributes:

        tracer_callbacks: The callback functions invoked in the trace. Mostly,
        it will consume the :class:`ActionMessageEntry` and produce the
        :class:`TraceMessageEntry` in the :class:`Q`

        source: The path to the executed file. To ensure the source is runnable,
        the ``PYTHONPATH`` must be set before running.

        args: The args used for the executed file.

    """
    tracer_callbacks: Optional[List[T_tracer_callback_func]] = None
    source: Optional[str] = None

    def __init__(self,
                 source: str,
                 args: List[str],
                 tracer_callbacks: Optional[List[T_tracer_callback_func]],
                 logger: LoggerLike):
        self.args = args
        self.tracer_callbacks = tracer_callbacks
        self.logger = logger
        if os.path.isabs(source):
            self.source = source
        else:
            for path in [os.curdir] + sys.path:
                if path is None:
                    continue
                f = os.path.join(path, source)
                try:
                    exists = os.path.exists(f)
                except UnicodeError:
                    exists = False
                if exists:
                    self.source = f
                    break

    def _build_module(self, source: str) -> ModuleType:
        mod_ty = ModuleType('__main__')
        mod_ty.__file__ = source
        mod_ty.__loader__ = DummyLoader
        mod_ty.__builtins__ = sys.modules['builtins']
        sys.modules['__main__'] = mod_ty
        return mod_ty

    def _build_args(self, args: List[str], source: str) -> None:
        self.origin_argv = sys.argv
        sys.argv = [source]
        for arg in args:
            sys.argv.append(arg)

    def _resume_args(self) -> None:
        sys.argv = self.origin_argv

    def _run(self, source: str, args: List[str]):
        source_byte = read_source_py(source)
        mod = self._build_module(source)
        self._build_args(args, source)
        try:
            code = compile(source_byte, source, "exec", dont_inherit=True)
        except SyntaxError:
            self.logger.error(f"Fail to compile the code: {source} due to the syntax error", exc_info=True)
            return
        except ValueError:
            self.logger.error(f"Fail to compile the code: {source} due to the inner null value", exc_info=True)
            return
        tracer = Tracer(self.tracer_callbacks, logger=self.logger)
        tracer.start()
        try:
            exec(code, mod.__dict__)
        except RuntimeError:
            self.logger.error(f"Fail to execute code: {source}", exc_info=True)
        finally:
            tracer.stop()
            self._resume_args()

    def run(self):
        assert self.source is not None
        self._run(self.source, self.args)



import json
import os.path
import re
import sys
from typing import Any, cast, Optional, List

from diagnose import fileutils
from diagnose.engine.manage import PluginManager
from diagnose.queues import TraceMessageEntry, Q
from diagnose.typings import T_frame, T_event, T_tracefunc, T_tracer_callback_func, LoggerLike


THIS_FILE = __file__.rstrip('co')


class Tracer:

    def __init__(self,
                 callbacks: Optional[List[T_tracer_callback_func]],
                 logger: LoggerLike):
        self.callbacks = callbacks
        self.logger = logger

    def _mangle_path(self, path: str) -> str:
        return os.path.abspath(path)

    def _is_inner_module(self, path: str) -> bool:
        if re.match(r'^<.*>$', path) is not None:
            return True
        return False

    def _trace_func(self, frame: T_frame, event: T_event, args: Any):
        sp = PluginManager.get_plugin('ScopePlugin')
        if frame.f_code.co_filename in THIS_FILE or self._is_inner_module(frame.f_code.co_filename) \
                or not sp.tracer_callback(frame, event, args):
            return None
        cb_rt = list()
        if self.callbacks is not None:
            for cb in self.callbacks:
                cb_rt.append({"plugin": cb(frame, event, args)}) if cb(frame, event, args) is not None else None
        self.logger.info(f"filename: {self._mangle_path(frame.f_code.co_filename)}, "
                         f"lineno: {frame.f_lineno}, cb_rt: {json.dumps(cb_rt)}")
        Q.put(TraceMessageEntry(0,
                                self._mangle_path(frame.f_code.co_filename),
                                frame.f_lineno, self.line_hash(frame), self.file_hash(frame), cb_rt))
        return self._trace_func

    def start(self):
        trace_func = sys.gettrace()
        if trace_func is not None:
            self.logger.warning("Trace function has been already amounted")
            sys.settrace(None)
        self._trace_func = cast(T_tracefunc, self._trace_func)
        self.logger.info("Trace function is amounted")
        sys.settrace(self._trace_func)

    def stop(self):
        sys.settrace(None)
        self.logger.info("Trace function is unamounted")

    def line_hash(self, frame: T_frame):
        return fileutils.generate_classname(os.path.abspath(frame.f_code.co_filename), int(frame.f_lineno))

    def file_hash(self, frame: T_frame):
        return fileutils.generate_classname(os.path.abspath(frame.f_code.co_filename))



import os.path
from typing import List, Optional

from diagnose import fileutils
from diagnose.engine.dispatch import Dispatcher
from diagnose.engine.report import Reporter
from diagnose.engine.run import PyRunner
from diagnose.fs import FS
from diagnose.fs import Path
from diagnose.logs import Logger
from diagnose.server import RenderServer
from diagnose.typings import LoggerLike
from diagnose.engine.manage import PluginManager


class Pipeline:

    def __init__(self,
                 source: str,
                 args: List[str],
                 scope_path: str,
                 *,
                 exclude_dir: Optional[List[str]] = None,
                 exclude_file: Optional[List[str]] = None,
                 max_workers: int = 2,
                 server_hostname: str = 'localhost',
                 port: int = 8765,
                 logger: Optional['LoggerLike'] = None):
        if logger is None:
            logger = Logger.get_logger('engine')
        self.source = source
        self.args = args
        self.exclude_dir = exclude_dir
        self.exclude_file = exclude_file
        self.logger = logger

        self.fs = FS(Path(scope_path),
                     exclude_dir=self.exclude_dir,
                     exclude_file=self.exclude_file)
        self.dispatcher = Dispatcher(max_workers=max_workers)
        self.render_server = RenderServer(hostname=server_hostname, port=port)
        self.reporter = Reporter(self.fs, os.path.join(fileutils.get_package_dir(), f'server{os.path.sep}templates'), logger=self.logger)

    def do_process(self):
        Pipeline.do_preprocess()
        self.logger.info("finish doing preprocess")

        runner = PyRunner(source=self.source,
                          args=self.args,
                          tracer_callbacks=[p.plugin.tracer_callback for _, p in PluginManager.plugins.items()
                                            if p.plugin and hasattr(p.plugin, 'tracer_callback')],
                          logger=self.logger)
        runner.run()
        Pipeline.do_postprocess()
        self.logger.info("finish doing postprocess")

    def do_server(self):
        self.render_server.run()

    def prepare(self):
        self.reporter.prepare()
        self.reporter.build_htmls()
        self.reporter.report()
        self.dispatcher.add_callable(self.do_process)
        self.dispatcher.add_callable(self.do_server)

    @classmethod
    def do_preprocess(cls):
        for name, disc in PluginManager.plugins.items():
            if disc.enable:
                assert hasattr(disc.plugin, 'on_preprocess')
                disc.plugin.on_preprocess()

    @classmethod
    def do_postprocess(cls):
        for name, disc in PluginManager.plugins.items():
            if disc.enable:
                assert hasattr(disc.plugin, 'on_postprocess')
                disc.plugin.on_postprocess()

    def run(self):
        self.prepare()
        self.dispatcher.dispatch()



import asyncio
import functools
from typing import Optional, Union, Sequence

from diagnose.logs import Logger
from diagnose.queues import Q
from diagnose.server.handler import WebsocketHandler
from diagnose.server.ws import serve

from diagnose.typings import LoggerLike


class RenderServer:

    def __init__(self,
                 hostname: Union[str, Sequence[str]],
                 port: int,
                 *,
                 logger: Optional['LoggerLike'] = None):
        self.running = False
        self.serve = None

        if logger is None:
            logger = Logger.get_logger('server')
        self.hostname = hostname
        self.port = port
        self.logger = logger

    def _build_serve(self):
        assert self.serve is None

        ws_handler = WebsocketHandler(queue=Q)

        self.serve = functools.partial(serve,
                                       ws_handler=ws_handler,
                                       host=self.hostname,
                                       port=self.port,
                                       logger=self.logger
                                       )

    def _run(self) -> serve:
        self.running = True
        if self.serve is None:
            self._build_serve()
        _serve = self.serve()
        self.loop = _serve.loop
        return _serve

    def run(self):
        async def _inner():
            async with self._run():
                await asyncio.Future()
        asyncio.run(_inner())



import os
import re
from os import PathLike
from typing import List, Tuple, Optional, Generator, Any, Callable

from diagnose import fileutils
from diagnose.logs import Logger
from diagnose.fs.models import Directory, File, Line
from diagnose.typings import LoggerLike


IS_PYCACHE = lambda d: d.endswith('__pycache__')


class FS:
    """
    A file system used for caching traversed :class:`File` or :class:`Directory` with the given path

    Attributes:
        path: The given path to be traversed.
        exclude_dirs: The excluded directories. It also support wildcard matches.
        exclude_files: The excluded files. It also support wildcard matches.
        logger: The logger for logging information.
        root: The root node of the file system. It can be a :class:`File` node or :class:`Directory` node.
    """

    def __init__(self,
                 path: PathLike | str,
                 exclude_dir: Optional[List[str]] = None,
                 exclude_file: Optional[List[str]] = None,
                 logger: Optional['LoggerLike'] = None):
        self.path = os.path.abspath(os.fspath(path))
        self.exclude_dirs = exclude_dir if exclude_dir is not None else []
        self.exclude_files = exclude_file if exclude_file is not None else []
        if logger is None:
            logger = Logger.get_logger('fs')
        self.logger = logger

        self.root = None
        self.build()

    # Copied from https://github.com/nedbat/coveragepy
    def _translate(self, pattern: str, case_sensitive=True):
        if not case_sensitive:
            pattern = pattern.lower()
        i, n = 0, len(pattern)
        res = []
        while i < n:
            c = pattern[i]
            i = i + 1
            if c == "*":
                res.append("[^/]*")
            elif c == "?":
                res.append(".")
            elif c == "[":
                j = i
                if j < n and pattern[j] == "!":
                    j = j + 1
                if j < n and pattern[j] == "]":
                    j = j + 1
                while j < n and pattern[j] != "]":
                    j = j + 1
                if j >= n:
                    res.append("\\[")
                else:
                    stuff = pattern[i:j].replace("\\", "\\\\")
                    i = j + 1
                    if stuff[0] == "!":
                        stuff = "^" + stuff[1:]
                    elif stuff[0] == "^":
                        stuff = "\\" + stuff
                    res.append("[%s]" % stuff)
            else:
                res.append(re.escape(c))
        return "".join(res)

    def match(self, pattern: str, name: str) -> bool:
        """
        Check if the given ``name`` can be matched with the ``pattern``. If it can be matched, the :meth:`match` method
        will return :obj:`True` , otherwise :obj:`False`.

        Args:
            pattern: matching pattern with ``*`` in it
            name: matched path

        Returns:
            if it can be matched
        """
        res = "(?ms)" + self._translate(pattern) + r"\Z"
        re_pat = re.compile(res)
        return re_pat.match(name) is not None

    def _scanf_dir(self, path: PathLike | str) -> Optional[Tuple[str, str]]:
        assert os.path.isabs(path)
        try:
            for info in os.listdir(path):
                yield path, info
        except OSError:
            self.logger.error(f'Fail to walk path {path}', exc_info=True)
            return None

    def _inspect_file(self, file: PathLike | str):
        assert os.path.isabs(file)
        file = os.fspath(file)
        _, ext = os.path.splitext(file)
        content = list()
        if ext == '.py' or ext == '.pyw':
            for (lineno, line) in fileutils.read_source_py_with_line(file):
                content.append(Line(lineno=lineno, content=str(line, encoding='UTF-8'), filename=file))
        else:
            try:
                content.append(
                    Line(lineno=1, content=str(fileutils.read_source(file), encoding='utf-8'), filename=file))
            except UnicodeDecodeError:
                self.logger.warning(f"Fail to inspect file {file}, which may be a binary file", exc_info=True)
        stat = fileutils.stat(file)
        return ext, content, stat[0], stat[1]

    def build(self):
        """
        Try to build the whole file system with inited ``path``, eg::

            fs = FS('..', exclude_dir=['.foo', '.bar'])
            fs.build('..')

        The method must be called before :meth:`walk` and :meth:`find`

        Returns:
            None
        """
        if os.path.isfile(self.path):
            f = File(filename=self.path, basename=os.path.basename(self.path))
            f.extension, f.lines, f.last_modified, f.size = self._inspect_file(self.path)
            self.root = f
            return f

        self.root = Directory(dirname=self.path, basename=os.path.basename(self.path), files_or_directories=[])
        self.root.last_modified, self.root.size = fileutils.stat(self.path)
        # The stack for memoizing Directory node. When a directory is found and its children nodes (sub-directory or
        # sub-file) are not scaned yet, it will be cached into the stack temporarily. Commonly, the nodes in the stack
        # for one time are in the same layer.
        stack = [
            (self.path, self.root)
        ]
        while len(stack) > 0:
            _parent_path, _dir = stack.pop(0)
            for _p, _d in self._scanf_dir(_parent_path):
                # XXX: Store the relative path for a while.
                _short_d = _d
                _d = os.path.join(_p, _d)
                if os.path.isdir(_d) and not IS_PYCACHE(_d):
                    self.logger.info(f"FS.build: Scanning the directory {_d}")
                    # Exclude the directories in `exclude_dirs`
                    if any([self.match(pattern, _d)
                            or self.match(pattern, _short_d)
                            for pattern in self.exclude_dirs]):
                        self.logger.info(f"FS.build: The directory {_d} is excluded")
                        continue
                    _new_dir = Directory(dirname=_d, basename=_short_d, files_or_directories=[])
                    _new_dir.last_modified, _new_dir.size = fileutils.stat(_d)
                    _dir.files_or_directories.append(_new_dir)
                    stack.append((_d, _new_dir))
                elif os.path.isfile(_d):
                    self.logger.info(f"FS.build: Scanning the file {_d}")
                    # Exclude the files in `exclude_files`
                    if any(([self.match(pattern, _d)
                             or self.match(pattern, _short_d)
                             for pattern in self.exclude_files])):
                        self.logger.info(f"FS.build: The file {_d} is excluded")
                        continue
                    _new_file = File(filename=_d, basename=os.path.basename(_d))
                    _new_file.extension, _new_file.lines, _new_file.last_modified, _new_file.size = self._inspect_file(_d)
                    _dir.files_or_directories.append(_new_file)
                else:
                    self.logger.warning(f"FS.build: Unexpected target {_d}")
                    continue

    def _walk(self, root_dir: Directory, hook: Callable[[File | Directory], Any]) -> Generator[File, Any, Any]:
        stack = [
            root_dir
        ]
        while len(stack) > 0:
            _d = stack.pop(0)
            if isinstance(_d, File):
                self.logger.info(f"FS.walk: Walking the file {_d}")
                hook(_d)
                yield _d
            elif isinstance(_d, Directory):
                hook(_d)
                self.logger.info(f"FS.walk: Walking the directory {_d}")
                for _e in _d.files_or_directories:
                    if isinstance(_e, File):
                        self.logger.info(f"FS.walk: Walking the file {_e}")
                        hook(_e)
                        yield _e
                    elif isinstance(_e, Directory):
                        stack.append(_e)
                    else:
                        raise RuntimeError(f'Unknown File in Directory {_d!r}')
            else:
                raise RuntimeError(f'Unknown File {_d!r}')

    def find(self, path: PathLike | str) -> Directory | File | None:
        """
        Try to find the :class:`Directory` or :class:`File` with given ``path``. If it exists, the method will return
        it, otherwise :obj:`None`.

        Note that this method must be invoked after :meth:`build`

        Args:
            path: The given path.

        Returns:
            The object the method found.

        Raises:
            RuntimeError: The object is neither :class:`File` nor :class:`Directory` RuntimeError: The :class:`FS` was
            not built.
        """
        path = os.fspath(path)
        path = os.path.abspath(path)
        stack = [
            self.root
        ]
        while len(stack) > 0:
            _d = stack.pop(0)
            if isinstance(_d, File):
                self.logger.info(f"FS.find: Finding the file {_d}")
                if _d.filename == path:
                    return _d
                else:
                    continue
            elif isinstance(_d, Directory):
                self.logger.info(f"FS.find: Finding the directory {_d}")
                if _d.dirname == path:
                    return _d
                else:
                    for _e in _d.files_or_directories:
                        stack.append(_e)
            else:
                raise RuntimeError(f'Unknown File {_d!r}')
        return None

    def _wrap_hook(self, hook: Optional[Callable[[File | Directory], None]] = None):
        return lambda x: hook(x) if hook is not None else None

    def walk(self,
             path: Optional[PathLike[str]] = None,
             hook: Optional[Callable[[File | Directory], None]] = None) -> Generator[File, Any, Any]:
        """
        Walk all :class:`Directory` or :class:`File` in the :class:`FS`.

        This method must be invoked after :meth:`build`::

            w = FS('..')
            w.build()
            for _w in w.walk('path\\to\\file'):
                print(_w)

        Args:
            path: The root directory for the walking if given.
            hook: The hook function invoked when walking to a :class:`File` or :class:`Directory`.

        Return:
            Files

        Raises:
            RuntimeError: The :class:`FS` was not built
            RuntimeError: The given path is not found
        """
        hook = self._wrap_hook(hook)
        if path is not None:
            res = self.find(path)
            if res is None:
                raise RuntimeError(f"Given file path {path} is not found")
            if isinstance(res, File):
                return res
            else:
                yield from self._walk(res, hook=hook)
        else:
            yield from self._walk(self.root, hook=hook)


if __name__ == '__main__':
    fs = FS('')
    for i in fs.walk():
        print(i)



import datetime
import io
import logging
from os import PathLike
from typing import ClassVar, Optional

__all__ = ['Logger']

#from typings import LoggerLike


class Logger(logging.Logger):
    """
    The logger implementation for the whole project.

    The :class:`Logger` is a subclass of :class:`logging.Logger` which is
    implemented by Vinay Sajip. This :class:`Logger` rewrites various logging
    methods and reorganizes :meth:`get_logger` and :meth:`redirect_to_file` to
    support the current situation and make it more object-oriented.

    The current :class:`Logger` is actually a dummy Logger and only provides the
    logging interface for users. The real :class:`Logger` is its class variable
    :obj:`_logger`.

    Attributes:
        _logger: The class variable for storing the real
            :class:`logging.Logger` which is appended to the Logger chain.
            See the `logging document`_ and `HOW-TO`_ for more details.

            .. _`logging document`: https://docs.python.org/3/library/logging.html
            .. _`HOW-TO`: https://docs.python.org/3/howto/logging.html

    """

    _logger: ClassVar[logging.Logger] = None

    @staticmethod
    def _acquire_lock():
        """
        Acquire the lock in the logging package.

        """
        getattr(logging, '_acquireLock')()

    @staticmethod
    def _release_lock():
        """
        Release the lock in the logging package.

        """
        getattr(logging, '_releaseLock')()
    
    @staticmethod
    def redirect_to_file(filename: PathLike,
                         mode: str = 'a',
                         logger: Optional['LoggerLike'] = None):
        """
        Redirect the output stream of :class:`Logger` to a file.

        The method :meth:`redirect_to_file` behaves like the function
        :func:`logging.basicConfig` in :mod:`logging`. Users can redirect the
        logging information to given file like this::

            Logger.redirect_to_file('foo.log')

        Note that the output of other Python program will not be affected. For
        example, the file in executed in ``exec``

        Args:
            filename(PathLike): The path of the redirected file.
            mode(str): The mode for processing files. See the `offical document`_
                for :mod:`os`.

                .. _`offical document`: https://docs.python.org/3/library/functions.html#open
            logger(Optional[LoggerLike]): The :class:`Logger` needs to redirect
                the output. If not provided, all :class:`Logger` will be
                redirected.

        """
        Logger._acquire_lock()
        # Copy from the logging module
        encoding = None
        errors = 'backslashreplace'
        try:
            if 'b' in mode:
                errors = None
            else:
                encoding = io.text_encoding(encoding)
            h = logging.FileHandler(filename, mode,
                                encoding=encoding, errors=errors)
            fmt = logging.Formatter(logging.BASIC_FORMAT, None, '%')
            h.setFormatter(fmt)
            if not logger:
                logging.root.addHandler(h)
            else:
                # XXX: Hack to the inner real Logger object
                logger._logger.addHandler(h)
                # Update the dict in logging module manually
                logging.Logger.manager.loggerDict.update([(logger.name.name, logger)])
        finally:
            Logger._release_lock()
    
    @classmethod
    def get_logger(cls, name=None) -> 'Logger':
        """
        Get the :class:`Logger` object with given name. If not provided, it will
        return the default :class:`Logger`.

        It behaves like the function :func:`logging.getLogger`.

        Args:
            name: The name of the :class:`Logger`. If it exists in the inner dict,
                it will returns directly. Otherwise, it will be initialized.

        Returns:
            Logger: The :class:`Logger` object.
        """
        cls._logger = logging.getLogger(name)
        # Real Logger must be wrapped so that it can invoke the methods bellow.
        return Logger(cls._logger)

    # Modify the methods in the logging package.

    def debug(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        _logger = self._logger
        if _logger.isEnabledFor(logging.DEBUG):
            msg = f'\033[38m[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]{msg}\033[0m'
            _logger._log(logging.DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.info("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        _logger = self._logger
        if _logger.isEnabledFor(logging.INFO):
            msg = f'\033[38m[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]{msg}\033[0m'
            _logger._log(logging.INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)
        """
        _logger = self._logger
        if _logger.isEnabledFor(logging.WARNING):
            msg = f'\033[33m[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]{msg}\033[0m'
            _logger._log(logging.WARNING, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.error("Houston, we have a %s", "major problem", exc_info=1)
        """
        _logger = self._logger
        if _logger.isEnabledFor(logging.ERROR):
            msg = f'\033[31m[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]{msg}\033[0m'
            _logger._log(logging.ERROR, msg, args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        """
        Convenience method for logging an ERROR with exception information.
        """
        self._logger.error(msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.critical("Houston, we have a %s", "major disaster", exc_info=1)
        """
        _logger = self._logger
        if _logger.isEnabledFor(logging.CRITICAL):
            msg = f'\033[31m[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]{msg}\033[0m'
            _logger._log(logging.CRITICAL, msg, args, **kwargs)

    def fatal(self, msg, *args, **kwargs):
        """
        Don't use this method, use critical() instead.
        """
        self._logger.critical(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """
        Log 'msg % args' with the integer severity 'level'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.log(level, "We have a %s", "mysterious problem", exc_info=1)
        """
        _logger = self._logger
        if not isinstance(level, int):
            raise TypeError("level must be an integer")

        if _logger.isEnabledFor(level):
            msg = f'\033[38m[{datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")}]{msg}\033[0m'
            _logger._log(level, msg, args, **kwargs)


"""
Copyright (c) Aymeric Augustin and contributors

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


import asyncio
import re
from typing import Any, Dict, Iterator, List, MutableMapping, Tuple

from diagnose.server.ws.exception import WebsocketException
from diagnose.server.ws.typings import HeadersLike

# Maximum total size of headers is around 128 * 8 KiB = 1 MiB.

MAX_HEADERS = 128

# Limit request line and header lines. 8KiB is the most common default
# configuration of popular HTTP servers.

MAX_LINE = 8192


# Regex for validating header names.

_token_re = re.compile(rb"[-!#$%&\'*+.^_`|~0-9a-zA-Z]+")

# Regex for validating header values.

_value_re = re.compile(rb"[\x09\x20-\x7e\x80-\xff]*")


class Headers(MutableMapping[str, str]):
    __slots__ = ["_dict", "_list"]

    # Like dict, Headers accepts an optional "mapping or iterable" argument.
    def __init__(self, *args: HeadersLike, **kwargs: str) -> None:
        self._dict: Dict[str, List[str]] = {}
        self._list: List[Tuple[str, str]] = []
        self.update(*args, **kwargs)

    def __str__(self) -> str:
        return "".join(f"{key}: {value}\r\n" for key, value in self._list) + "\r\n"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._list!r})"

    def copy(self) -> 'Headers':
        copy = self.__class__()
        copy._dict = self._dict.copy()
        copy._list = self._list.copy()
        return copy

    def serialize(self) -> bytes:
        # Since headers only contain ASCII characters, we can keep this simple.
        return str(self).encode()

    # Collection methods

    def __contains__(self, key: object) -> bool:
        return isinstance(key, str) and key.lower() in self._dict

    def __iter__(self) -> Iterator[str]:
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    # MutableMapping methods

    def __getitem__(self, key: str) -> str:
        value = self._dict[key.lower()]
        if len(value) == 1:
            return value[0]
        else:
            raise RuntimeError(f"Multiple key found {key}")

    def __setitem__(self, key: str, value: str) -> None:
        self._dict.setdefault(key.lower(), []).append(value)
        self._list.append((key, value))

    def __delitem__(self, key: str) -> None:
        key_lower = key.lower()
        self._dict.__delitem__(key_lower)
        # This is inefficient. Fortunately deleting HTTP headers is uncommon.
        self._list = [(k, v) for k, v in self._list if k.lower() != key_lower]

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Headers):
            return NotImplemented
        return self._dict == other._dict

    def clear(self) -> None:
        """
        Remove all headers.

        """
        self._dict = {}
        self._list = []

    def update(self, *args: 'HeadersLike', **kwargs: str) -> None:
        """
        Update from a :class:`Headers` instance and/or keyword arguments.

        """
        args = tuple(
            arg.raw_items() if isinstance(arg, Headers) else arg for arg in args
        )
        super().update(*args, **kwargs)

    # Methods for handling multiple values

    def get_all(self, key: str) -> List[str]:
        """
        Return the (possibly empty) list of all values for a header.

        Args:
            key: header name.

        """
        return self._dict.get(key.lower(), [])

    def raw_items(self) -> Iterator[Tuple[str, str]]:
        """
        Return an iterator of all values as ``(name, value)`` pairs.

        """
        return iter(self._list)


def d(value: bytes) -> str:
    """
    Decode a bytestring for interpolating into an error message.

    """
    return value.decode(errors="backslashreplace")


async def read_request(stream: asyncio.StreamReader) -> Tuple[str, Headers]:
    """
    Read an HTTP/1.1 GET request and return ``(path, headers)``.

    ``path`` isn't URL-decoded or validated in any way.

    ``path`` and ``headers`` are expected to contain only ASCII characters.
    Other characters are represented with surrogate escapes.

    :func:`read_request` doesn't attempt to read the request body because
    WebSocket handshake requests don't have one. If the request contains a
    body, it may be read from ``stream`` after this coroutine returns.

    Args:
        stream: input to read the request from

    Raises:
        EOFError: if the connection is closed without a full HTTP request
        SecurityError: if the request exceeds a security limit
        ValueError: if the request isn't well formatted

    """
    # https://www.rfc-editor.org/rfc/rfc7230.html#section-3.1.1

    # Parsing is simple because fixed values are expected for method and
    # version and because path isn't checked. Since WebSocket software tends
    # to implement HTTP/1.1 strictly, there's little need for lenient parsing.

    try:
        request_line = await read_line(stream)
    except EOFError as exc:
        raise EOFError("connection closed while reading HTTP request line") from exc

    try:
        method, raw_path, version = request_line.split(b" ", 2)
    except ValueError:  # not enough values to unpack (expected 3, got 1-2)
        raise ValueError(f"invalid HTTP request line: {d(request_line)}") from None

    if method != b"GET":
        raise ValueError(f"unsupported HTTP method: {d(method)}")
    if version != b"HTTP/1.1":
        raise ValueError(f"unsupported HTTP version: {d(version)}")
    path = raw_path.decode("ascii", "surrogateescape")

    headers = await read_headers(stream)

    return path, headers


async def read_response(stream: asyncio.StreamReader) -> Tuple[int, str, Headers]:
    """
    Read an HTTP/1.1 response and return ``(status_code, reason, headers)``.

    ``reason`` and ``headers`` are expected to contain only ASCII characters.
    Other characters are represented with surrogate escapes.

    :func:`read_request` doesn't attempt to read the response body because
    WebSocket handshake responses don't have one. If the response contains a
    body, it may be read from ``stream`` after this coroutine returns.

    Args:
        stream: input to read the response from

    Raises:
        EOFError: if the connection is closed without a full HTTP response
        SecurityError: if the response exceeds a security limit
        ValueError: if the response isn't well formatted

    """
    # https://www.rfc-editor.org/rfc/rfc7230.html#section-3.1.2

    # As in read_request, parsing is simple because a fixed value is expected
    # for version, status_code is a 3-digit number, and reason can be ignored.

    try:
        status_line = await read_line(stream)
    except EOFError as exc:
        raise EOFError("connection closed while reading HTTP status line") from exc

    try:
        version, raw_status_code, raw_reason = status_line.split(b" ", 2)
    except ValueError:  # not enough values to unpack (expected 3, got 1-2)
        raise ValueError(f"invalid HTTP status line: {d(status_line)}") from None

    if version != b"HTTP/1.1":
        raise ValueError(f"unsupported HTTP version: {d(version)}")
    try:
        status_code = int(raw_status_code)
    except ValueError:  # invalid literal for int() with base 10
        raise ValueError(f"invalid HTTP status code: {d(raw_status_code)}") from None
    if not 100 <= status_code < 1000:
        raise ValueError(f"unsupported HTTP status code: {d(raw_status_code)}")
    if not _value_re.fullmatch(raw_reason):
        raise ValueError(f"invalid HTTP reason phrase: {d(raw_reason)}")
    reason = raw_reason.decode()

    headers = await read_headers(stream)

    return status_code, reason, headers


async def read_headers(stream: asyncio.StreamReader) -> Headers:
    """
    Read HTTP headers from ``stream``.

    Non-ASCII characters are represented with surrogate escapes.

    """
    # https://www.rfc-editor.org/rfc/rfc7230.html#section-3.2

    # We don't attempt to support obsolete line folding.

    headers = Headers()
    for _ in range(MAX_HEADERS + 1):
        try:
            line = await read_line(stream)
        except EOFError as exc:
            raise EOFError("connection closed while reading HTTP headers") from exc
        if line == b"":
            break

        try:
            raw_name, raw_value = line.split(b":", 1)
        except ValueError:  # not enough values to unpack (expected 2, got 1)
            raise ValueError(f"invalid HTTP header line: {d(line)}") from None
        if not _token_re.fullmatch(raw_name):
            raise ValueError(f"invalid HTTP header name: {d(raw_name)}")
        raw_value = raw_value.strip(b" \t")
        if not _value_re.fullmatch(raw_value):
            raise ValueError(f"invalid HTTP header value: {d(raw_value)}")

        name = raw_name.decode("ascii")  # guaranteed to be ASCII at this point
        value = raw_value.decode("ascii", "surrogateescape")
        headers[name] = value

    else:
        raise WebsocketException("SecurityError too many HTTP headers")

    return headers


async def read_line(stream: asyncio.StreamReader) -> bytes:
    """
    Read a single line from ``stream``.

    CRLF is stripped from the return value.

    """
    # Security: this is bounded by the StreamReader's limit (default = 32 KiB).
    line = await stream.readline()
    # Security: this guarantees header values are small (hard-coded = 8 KiB)
    if len(line) > MAX_LINE:
        raise WebsocketException("SecurityError line too long")
    # Not mandatory but safe - https://www.rfc-editor.org/rfc/rfc7230.html#section-3.5
    if not line.endswith(b"\r\n"):
        raise EOFError("line without CRLF")
    return line[:-2]


from setuptools import setup, find_packages


def get_version():
    f = open('./VERSION', 'r', encoding='utf-8')
    version = f.readline().strip()
    f.close()
    return version


def get_long_descript():
    f = open('./README.rst', 'r', encoding='utf-8')
    long_descript = f.read()
    f.close()
    return long_descript


if __name__ == '__main__':
    setup(name='hot-diagnose',
          version=get_version(),
          author='littlebutt',
          author_email='luogan1996@icloud.com',
          description="The runtime code diagnose tool",
          long_description=get_long_descript(),
          url='https://github.com/littlebutt/hot-diagnose',
          python_requires='>=3.10',
          packages=find_packages(),
          include_package_data=True)


import concurrent.futures
from concurrent.futures import Future
from typing import Callable, Any, Dict


class Dispatcher:

    def __init__(self,
                 max_workers: int = 2,):
        self.pool_executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        self.callable_group = list()
        self.future_map: Dict[str, Future] = dict()

    def add_callable(self, _callable: Callable[[...], Any]):
        self.callable_group.append(_callable)

    def dispatch(self):
        with self.pool_executor as executor:
            for cb in self.callable_group:
                self.future_map.update({cb.__name__: executor.submit(cb)})



import os
from os import PathLike
from typing import Optional, cast, List, Any

from diagnose import fileutils
from diagnose.fs import FS, File, Directory
from diagnose.logs import Logger
from diagnose.server import Template
from diagnose.typings import LoggerLike


class Reporter:

    def __init__(self,
                 fs: FS,
                 templates_dir: PathLike,
                 *,
                 logger: Optional['LoggerLike'] = None):
        self.fs = fs
        self.templates_dir = os.fspath(templates_dir)
        if logger is None:
            logger = Logger.get_logger('engine')
        self.logger = logger

        # The home directory for storing HTML files.
        self.root_dir = fileutils.mkdir(fileutils.get_home_dir(), '.hot-diagnose')

        # The dict of templates infomation whose key is the templates' name and the value is the templates' text.
        self.template_dict = dict()

        # The dict of file context whose key is the HTML files' name and the value is its context dict. The HTML files
        # are generated according to tamplates above while the context is generated by executed files.
        self.template_context = dict({
                'Directories': [],
                'Files': []
            })

    def prepare(self):
        fs = FS(self.templates_dir, logger=self.logger)

        def _append_dict(target):
            if isinstance(target, File):
                target = cast(File, target)
                self.template_dict[target.basename] = str(target)

        for _ in fs.walk(hook=_append_dict):
            pass

    def _walk_hook(self, target: File | Directory) -> None:
        if isinstance(target, File):
            target = cast(File, target)
            self.template_context['Files'].append(target)
        elif isinstance(target, Directory):
            target = cast(Directory, target)
            self.template_context['Directories'].append(target)
        else:
            self.logger.warning(f"Unexpected file {target}")

    def build_htmls(self):
        for _ in self.fs.walk(hook=self._walk_hook):
            pass

        self.template_context.update({'escape': escape, 'len': len, 'is_file': is_file})

        template = Template(self.template_dict['index.html'], self.template_context)
        fileutils.write_file(os.path.join(self.root_dir, 'index.html'), template.render())
        fileutils.write_file(os.path.join(self.root_dir, 'index.css'), self.template_dict['index.css'])
        fileutils.write_file(os.path.join(self.root_dir, 'control.js'), self.template_dict['control.js'])

    def report(self):
        import webbrowser
        webbrowser.open(os.path.join(self.root_dir, 'index.html'))


def escape(line: str):
    return line.replace("&", "&amp;").replace(" ", "&nbsp;").replace("<", "&lt;")


def is_file(content: File | Directory):
    return isinstance(content, File)



"""
Copyright (c) Aymeric Augustin and contributors

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of the copyright holder nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import asyncio
import codecs
import collections
import email
import http
import logging
import ssl
import struct
import time
import random
from typing import Callable, Awaitable, Any, Optional, cast, Tuple, List, Deque, Dict, AsyncIterator, AsyncIterable, \
    Iterable, Mapping, Union

from diagnose.server.ws.misc import State, build_response, check_request
from diagnose.server.ws.exception import WebsocketException
from diagnose.server.ws.frames import Close, OP_CLOSE, Frame, Opcode, OK_CLOSE_CODES, OP_TEXT, OP_BINARY, OP_PING, OP_PONG, \
    OP_CONT, prepare_ctrl, prepare_data
from diagnose.server.ws.http11 import Headers, read_request
from diagnose.server.ws.typings import Data, ServerLike
from diagnose.typings import LoggerLike


class WebSocketServerProtocol(asyncio.Protocol):
    """
    WebSocket server connection.

    :class:`WebSocketServerProtocol` provides :meth:`recv` and :meth:`send`
    coroutines for receiving and sending messages.

    It supports asynchronous iteration to receive messages::

        async for message in websocket:
            await process(message)

    The iterator exits normally when the connection is closed with close code
    1000 (OK) or 1001 (going away) or without a close code. It raises
    a :exc:`WebsocketException` when the connection
    is closed with any other code.

    Args:
        ws_server: WebSocket server that created this connection.
    """

    def __init__(
            self,
            ws_handler: Callable[['WebSocketServerProtocol'], Awaitable[Any]],
            ws_server: ServerLike,
            *,
            read_limit: int = 2 ** 16,
            write_limit: int = 2 ** 256,
            max_size: Optional[int] = 2 ** 20,
            max_queue: Optional[int] = 2 ** 5,
            loop: Optional[asyncio.AbstractEventLoop] = None,
            close_timeout=10,
            logger: Optional['LoggerLike'] = None,
            ping_interval: Optional[float] = 20,
            ping_timeout: Optional[float] = 20,
            **kwargs: Any,
    ) -> None:
        if logger is None:
            logger = logging.getLogger("websockets.server")
        self.ws_handler = ws_handler
        self.ws_server = ws_server
        self.read_limit = read_limit
        self.write_limit = write_limit
        self.max_size = max_size
        self.max_queue = max_queue
        self.loop = loop
        self.close_timeout = close_timeout
        self.connection_lost_waiter: asyncio.Future[None] = loop.create_future()
        self.reader = asyncio.StreamReader(limit=read_limit // 2, loop=loop)
        self.debug = True
        self.logger = logger
        self.ping_interval = ping_interval
        self.ping_timeout = ping_timeout

        self.state = State.CONNECTING

        # Protect sending fragmented messages.
        self._fragmented_message_waiter: Optional[asyncio.Future[None]] = None

        self.close_rcvd: Optional[Close] = None
        self.close_sent: Optional[Close] = None
        self.close_rcvd_then_sent: Optional[bool] = None

        self._paused = False
        self._drain_waiter: Optional[asyncio.Future[None]] = None
        self._drain_lock = asyncio.Lock()

        # Mapping of ping IDs to pong waiters, in chronological order.
        self.pings: Dict[bytes, Tuple[asyncio.Future[float], float]] = {}

        # Task running the data transfer.
        self.transfer_data_task: asyncio.Task[None]

        # Exception that occurred during data transfer, if any.
        self.transfer_data_exc: Optional[BaseException] = None

        # Task sending keepalive pings.
        self.keepalive_ping_task: asyncio.Task[None]

        # Task closing the TCP connection.
        self.close_connection_task: asyncio.Task[None]

        # Queue of received messages.
        self.messages: Deque[Data] = collections.deque()
        self._pop_message_waiter: Optional[asyncio.Future[None]] = None
        self._put_message_waiter: Optional[asyncio.Future[None]] = None

    @property
    def open(self) -> bool:
        """
        :obj:`True` when the connection is open; :obj:`False` otherwise.

        This attribute may be used to detect disconnections. However, this
        approach is discouraged per the EAFP_ principle. Instead, you should
        handle :exc:`WebsocketException` exceptions.

        .. _EAFP: https://docs.python.org/3/glossary.html#term-eafp

        """
        return self.state is State.OPEN and not self.transfer_data_task.done()

    @property
    def closed(self) -> bool:
        """
        :obj:`True` when the connection is closed; :obj:`False` otherwise.

        Be aware that both :attr:`open` and :attr:`closed` are :obj:`False`
        during the opening and closing sequences.

        """
        return self.state is State.CLOSED

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """
        Register connection and initialize a task to handle it.

        """
        transport = cast(asyncio.Transport, transport)
        transport.set_write_buffer_limits(self.write_limit)
        self.transport = transport

        # Copied from asyncio.StreamReaderProtocol
        self.reader.set_transport(transport)
        # Register the connection with the server before creating the handler
        # task. Registering at the beginning of the handler coroutine would
        # create a race condition between the creation of the task, which
        # schedules its execution, and the moment the handler starts running.
        self.ws_server.register(self)
        self.handler_task = self.loop.create_task(self.handler())

    async def wait_for_connection_lost(self) -> bool:
        """
        Wait until the TCP connection is closed or ``self.close_timeout`` elapses.

        Return :obj:`True` if the connection is closed and :obj:`False`
        otherwise.

        """
        if not self.connection_lost_waiter.done():
            try:
                await asyncio.wait_for(
                    asyncio.shield(self.connection_lost_waiter),
                    self.close_timeout,
                )
            except asyncio.TimeoutError:
                pass
        # Re-check self.connection_lost_waiter.done() synchronously because
        # connection_lost() could run between the moment the timeout occurs
        # and the moment this coroutine resumes running.
        return self.connection_lost_waiter.done()

    async def close_transport(self) -> None:
        """
        Close the TCP connection.

        """
        # If connection_lost() was called, the TCP connection is closed.
        # However, if TLS is enabled, the transport still needs closing.
        # Else asyncio complains: ResourceWarning: unclosed transport.
        if self.connection_lost_waiter.done() and self.transport.is_closing():
            return

        # Close the TCP connection. Buffers are flushed asynchronously.
        if self.debug:
            self.logger.debug("x closing TCP connection")
        self.transport.close()

        if await self.wait_for_connection_lost():
            return
        if self.debug:
            self.logger.debug("! timed out waiting for TCP close")

        # Abort the TCP connection. Buffers are discarded.
        if self.debug:
            self.logger.debug("x aborting TCP connection")
        self.transport.abort()

        # connection_lost() is called quickly after aborting.
        # Coverage marks this line as a partially executed branch.
        # I suspect a bug in coverage. Ignore it for now.
        await self.wait_for_connection_lost()  # pragma: no cover

    def write_frame_sync(self, fin: bool, opcode: int, data: bytes) -> None:
        frame = Frame(Opcode(opcode), data, fin)
        if self.debug:
            self.logger.debug(f"> {frame}")
        frame.write(
            self.transport.write,
            mask=False,
        )

    # Copied from asyncio.FlowControlMixin
    async def _drain_helper(self) -> None:
        if self.connection_lost_waiter.done():
            raise ConnectionResetError("Connection lost")
        # if not self._paused:
        #     return
        # waiter = self._drain_waiter
        # assert waiter is None or waiter.cancelled()
        # waiter = self.loop.create_future()
        # self._drain_waiter = waiter
        # await waiter

    # Copied from asyncio.StreamWriter
    async def _drain(self) -> None:
        if self.reader is not None:
            exc = self.reader.exception()
            if exc is not None:
                raise exc
        if self.transport is not None:
            if self.transport.is_closing():
                # Yield to the event loop so connection_lost() may be
                # called.  Without this, _drain_helper() would return
                # immediately, and code that calls
                #     write(...); yield from drain()
                # in a loop would never call connection_lost(), so it
                # would not see an error when the socket is closed.
                await asyncio.sleep(0)
        await self._drain_helper()

    async def close_connection(self) -> None:
        """
        When the opening handshake succeeds, :meth:`connection_open` starts
        this coroutine in a task. It waits for the data transfer phase to
        complete then it closes the TCP connection cleanly.

        When the opening handshake fails, :meth:`fail_connection` does the
        same. There's no data transfer phase in that case.
        """
        try:
            # Wait for the data transfer phase to complete.
            if hasattr(self, "transfer_data_task"):
                try:
                    await self.transfer_data_task
                except asyncio.CancelledError:
                    pass

            # Cancel the keepalive ping task.
            if hasattr(self, "keepalive_ping_task"):
                self.keepalive_ping_task.cancel()

            # Half-close the TCP connection if possible (when there's no TLS).
            if self.transport.can_write_eof():
                if self.debug:
                    self.logger.debug("x half-closing TCP connection")
                # write_eof() doesn't document which exceptions it raises.
                # "[Errno 107] Transport endpoint is not connected" happens,
                # but it isn't completely clear under which circumstances.
                # uvloop can raise RuntimeError here.
                try:
                    self.transport.write_eof()
                except (OSError, RuntimeError):  # pragma: no cover
                    pass

                if await self.wait_for_connection_lost():
                    return  # pragma: no cover
                if self.debug:
                    self.logger.debug("! timed out waiting for TCP close")

        finally:
            # The try/finally ensures that the transport never remains open,
            # even if this coroutine is canceled (for example).
            await self.close_transport()

    def fail_connection(self, code: int = 1006, reason: str = "") -> None:
        """
        Fail the WebSocket Connection
        """
        if self.debug:
            self.logger.debug(f"! failing connection with code {code}")

        # Cancel transfer_data_task if the opening handshake succeeded.
        # cancel() is idempotent and ignored if the task is done already.
        if hasattr(self, "transfer_data_task"):
            self.transfer_data_task.cancel()

        # Send a close frame when the state is OPEN (a close frame was already
        # sent if it's CLOSING), except when failing the connection because of
        # an error reading from or writing to the network.
        # Don't send a close frame if the connection is broken.
        if code != 1006 and self.state is State.OPEN:
            close = Close(code, reason)

            # Write the close frame without draining the write buffer.

            # Keeping fail_connection() synchronous guarantees it can't
            # get stuck and simplifies the implementation of the callers.
            # Not drainig the write buffer is acceptable in this context.

            # This duplicates a few lines of code from write_close_frame().

            self.state = State.CLOSING
            if self.debug:
                self.logger.debug("= connection is CLOSING")

            # If self.close_rcvd was set, the connection state would be
            # CLOSING. Therefore, self.close_rcvd isn't set, and we don't
            # have to set self.close_rcvd_then_sent.
            assert self.close_rcvd is None
            self.close_sent = close

            self.write_frame_sync(True, OP_CLOSE, close.serialize())

        # Start close_connection_task if the opening handshake didn't succeed.
        if not hasattr(self, "close_connection_task"):
            self.close_connection_task = self.loop.create_task(self.close_connection())

    def connection_closed_exc(self) -> WebsocketException:
        exc: WebsocketException
        if (
                self.close_rcvd is not None
                and self.close_rcvd.code in OK_CLOSE_CODES
                and self.close_sent is not None
                and self.close_sent.code in OK_CLOSE_CODES
        ):
            exc = WebsocketException(
                f"ConnectionClosedOK: {self.close_rcvd} {self.close_sent} {self.close_rcvd_then_sent}")
        else:
            exc = WebsocketException(
                f"ConnectionClosedError: {self.close_rcvd} {self.close_sent} {self.close_rcvd_then_sent}")
        # Chain to the exception that terminated data transfer, if any.
        exc.__cause__ = self.transfer_data_exc
        return exc

    async def ensure_open(self) -> None:
        """
        Check that the WebSocket connection is open.

        Raise :exc:`WebsocketException` if it isn't.

        """
        # Handle cases from most common to least common for performance.
        if self.state is State.OPEN:
            # If self.transfer_data_task exited without a closing handshake,
            # self.close_connection_task may be closing the connection, going
            # straight from OPEN to CLOSED.
            if self.transfer_data_task.done():
                await asyncio.shield(self.close_connection_task)
                raise self.connection_closed_exc()
            else:
                return

        if self.state is State.CLOSED:
            raise self.connection_closed_exc()

        if self.state is State.CLOSING:
            # If we started the closing handshake, wait for its completion to
            # get the proper close code and reason. self.close_connection_task
            # will complete within 4 or 5 * close_timeout after close(). The
            # CLOSING state also occurs when failing the connection. In that
            # case self.close_connection_task will complete even faster.
            await asyncio.shield(self.close_connection_task)
            raise self.connection_closed_exc()

        # Control may only reach this point in buggy third-party subclasses.
        assert self.state is State.CONNECTING
        raise WebsocketException("InvalidState: WebSocket connection isn't established yet")

    async def drain(self) -> None:
        try:
            async with self._drain_lock:
                # Handle flow control automatically.
                await self._drain()
        except ConnectionError:
            # Terminate the connection if the socket died.
            self.fail_connection()
            # Wait until the connection is closed to raise ConnectionClosed
            # with the correct code and reason.
            await self.ensure_open()

    async def write_frame(
            self, fin: bool, opcode: int, data: bytes, *, _state: int = State.OPEN
    ) -> None:
        # Defensive assertion for protocol compliance.
        if self.state is not _state:
            raise WebsocketException(f"InvalidState: Cannot write to a WebSocket in the {self.state.name} state")
        self.write_frame_sync(fin, opcode, data)
        await self.drain()

    async def write_close_frame(
            self, close: Close, data: Optional[bytes] = None
    ) -> None:
        """
        Write a close frame if and only if the connection state is OPEN.

        This dedicated coroutine must be used for writing close frames to
        ensure that at most one close frame is sent on a given connection.

        """
        # Test and set the connection state before sending the close frame to
        # avoid sending two frames in case of concurrent calls.
        if self.state is State.OPEN:
            # 7.1.3. The WebSocket Closing Handshake is Started
            self.state = State.CLOSING
            if self.debug:
                self.logger.debug("= connection is CLOSING")

            self.close_sent = close
            if self.close_rcvd is not None:
                self.close_rcvd_then_sent = True
            if data is None:
                data = close.serialize()

            # 7.1.2. Start the WebSocket Closing Handshake
            await self.write_frame(True, OP_CLOSE, data, _state=State.CLOSING)

    async def close(self, code: int = 1000, reason: str = "") -> None:
        """
        Perform the closing handshake.

        :meth:`close` waits for the other end to complete the handshake and
        for the TCP connection to terminate. As a consequence, there's no need
        to await :meth:`wait_closed` after :meth:`close`.

        :meth:`close` is idempotent: it doesn't do anything once the
        connection is closed.

        Wrapping :func:`close` in :func:`~asyncio.create_task` is safe, given
        that errors during connection termination aren't particularly useful.

        Canceling :meth:`close` is discouraged. If it takes too long, you can
        set a shorter ``close_timeout``. If you don't want to wait, let the
        Python process exit, then the OS will take care of closing the TCP
        connection.

        Args:
            code: WebSocket close code.
            reason: WebSocket close reason.

        """
        try:
            await asyncio.wait_for(
                self.write_close_frame(Close(code, reason)),
                self.close_timeout
            )
        except asyncio.TimeoutError:
            # If the close frame cannot be sent because the send buffers
            # are full, the closing handshake won't complete anyway.
            # Fail the connection to shut down faster.
            self.fail_connection()

        # If no close frame is received within the timeout, wait_for() cancels
        # the data transfer task and raises TimeoutError.

        # If close() is called multiple times concurrently and one of these
        # calls hits the timeout, the data transfer task will be canceled.
        # Other calls will receive a CancelledError here.

        try:
            # If close() is canceled during the wait, self.transfer_data_task
            # is canceled before the timeout elapses.
            await asyncio.wait_for(
                self.transfer_data_task,
                self.close_timeout,
            )
        except (asyncio.TimeoutError, asyncio.CancelledError):
            pass

        # Wait for the close connection task to close the TCP connection.
        await asyncio.shield(self.close_connection_task)

    async def handler(self) -> None:
        """
        Handle the lifecycle of a WebSocket connection.

        Since this method doesn't have a caller able to handle exceptions, it
        attempts to log relevant ones and guarantees that the TCP connection is
        closed before exiting.

        """
        try:

            try:
                await self.handshake()
            except asyncio.CancelledError:
                raise
            except ConnectionError:
                raise
            except WebsocketException as exc:
                if exc.type == 'AbortHandshake':
                    status, headers, body = None, None, None
                elif exc.type == 'InvalidUpgrade':
                    if self.debug:
                        self.logger.debug(f"! invalid upgrade {exc}")
                    status, headers, body = (
                        http.HTTPStatus.UPGRADE_REQUIRED,
                        Headers([("Upgrade", "websocket")]),
                        (
                            f"Failed to open a WebSocket connection: {exc}.\n"
                            f"\n"
                            f"You cannot access a WebSocket server directly "
                            f"with a browser. You need a WebSocket client.\n"
                        ).encode(),
                    )
                elif exc.type == 'InvalidHandshake':
                    if self.debug:
                        self.logger.debug(f"! invalid handshake {exc}")
                    status, headers, body = (
                        http.HTTPStatus.BAD_REQUEST,
                        Headers(),
                        f"Failed to open a WebSocket connection: {exc}.\n".encode(),
                    )
                else:
                    self.logger.error(f"opening handshake failed {exc}")
                    status, headers, body = (
                        http.HTTPStatus.INTERNAL_SERVER_ERROR,
                        Headers(),
                        (
                            b"Failed to open a WebSocket connection.\n"
                            b"See server log for more information.\n"
                        ),
                    )

                headers.setdefault("Date", email.utils.formatdate(usegmt=True))

                headers.setdefault("Content-Length", str(len(body)))
                headers.setdefault("Content-Type", "text/plain")
                headers.setdefault("Connection", "close")

                self.write_http_response(status, headers, body)
                self.logger.info(
                    f"connection failed ({status.value} {status.phrase})"
                )
                await self.close_transport()
                return

            try:
                await self.ws_handler(self)
            except Exception as exc:
                self.logger.error(f"connection handler failed {exc}")
                if not self.closed:
                    self.fail_connection(1011)
                raise

            try:
                await self.close()
            except ConnectionError:
                raise
            except Exception as exc:
                self.logger.error(f"closing handshake failed {exc}")
                raise

        except Exception:
            # Last-ditch attempt to avoid leaking connections on errors.
            try:
                self.transport.close()
            except Exception:
                pass

        finally:
            # Unregister the connection with the server when the handler task
            # terminates. Registration is tied to the lifecycle of the handler
            # task because the server waits for tasks attached to registered
            # connections before terminating.
            self.ws_server.unregister(self)
            self.logger.info("connection closed")

    async def read_http_request(self) -> Tuple[str, Headers]:
        """
        Read request line and headers from the HTTP request.

        If the request contains a body, it may be read from ``self.reader``
        after this coroutine returns.

        Raises:
            WebsocketException: if the HTTP message is malformed or isn't an
                HTTP/1.1 GET request.

        """
        try:
            path, headers = await read_request(self.reader)
        except asyncio.CancelledError:
            raise
        except Exception as exc:
            raise WebsocketException("InvalidMessage did not receive a valid HTTP request") from exc

        if self.debug:
            self.logger.debug("< GET %s HTTP/1.1", path)
            for key, value in headers.raw_items():
                self.logger.debug("< %s: %s", key, value)

        self.path = path
        self.request_headers = headers

        return path, headers

    def write_http_response(
            self, status: http.HTTPStatus, headers: Headers, body: Optional[bytes] = None
    ) -> None:
        """
        Write status line and headers to the HTTP response.

        This coroutine is also able to write a response body.

        """
        self.response_headers = headers

        if self.debug:
            self.logger.debug("> HTTP/1.1 %d %s", status.value, status.phrase)
            for key, value in headers.raw_items():
                self.logger.debug("> %s: %s", key, value)
            if body is not None:
                self.logger.debug("> [body] (%d bytes)", len(body))

        # Since the status line and headers only contain ASCII characters,
        # we can keep this simple.
        response = f"HTTP/1.1 {status.value} {status.phrase}\r\n"
        response += str(headers)

        self.transport.write(response.encode())

        if body is not None:
            self.transport.write(body)

    async def read_frame(self, max_size: Optional[int]) -> Frame:
        """
        Read a single frame from the connection.

        """
        frame = await Frame.read(
            self.reader.readexactly,
            mask=True,
            max_size=max_size
        )
        if self.debug:
            self.logger.debug(f"< {frame}")
        return frame

    async def pong(self, data: Data = b"") -> None:
        """
        Send a Pong_.

        .. _Pong: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.5.3

        An unsolicited pong may serve as a unidirectional heartbeat.

        Canceling :meth:`pong` is discouraged. If :meth:`pong` doesn't return
        immediately, it means the write buffer is full. If you don't want to
        wait, you should close the connection.

        Args:
            data (Data): payload of the pong; a string will be encoded to
                UTF-8.

        Raises:
            WebsocketException: when the connection is closed.

        """
        await self.ensure_open()

        data = prepare_ctrl(data)

        await self.write_frame(True, OP_PONG, data)

    async def ping(self, data: Optional[Data] = None) -> Awaitable[None]:
        """
        Send a Ping_.

        .. _Ping: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.5.2

        A ping may serve as a keepalive, as a check that the remote endpoint
        received all messages up to this point, or to measure :attr:`latency`.

        Canceling :meth:`ping` is discouraged. If :meth:`ping` doesn't return
        immediately, it means the write buffer is full. If you don't want to
        wait, you should close the connection.

        Canceling the :class:`~asyncio.Future` returned by :meth:`ping` has no
        effect.

        Args:
            data (Optional[Data]): payload of the ping; a string will be
                encoded to UTF-8; or :obj:`None` to generate a payload
                containing four random bytes.

        Returns:
            ~asyncio.Future[float]: A future that will be completed when the
            corresponding pong is received. You can ignore it if you don't
            intend to wait. The result of the future is the latency of the
            connection in seconds.

            ::

                pong_waiter = await ws.ping()
                # only if you want to wait for the corresponding pong
                latency = await pong_waiter

        Raises:
            WebsocketException: when the connection is closed.
            RuntimeError: if another ping was sent with the same data and
                the corresponding pong wasn't received yet.

        """
        await self.ensure_open()

        if data is not None:
            data = prepare_ctrl(data)

        # Protect against duplicates if a payload is explicitly set.
        if data in self.pings:
            raise RuntimeError("already waiting for a pong with the same data")

        # Generate a unique random payload otherwise.
        while data is None or data in self.pings:
            data = struct.pack("!I", random.getrandbits(32))

        pong_waiter = self.loop.create_future()
        # Resolution of time.monotonic() may be too low on Windows.
        ping_timestamp = time.perf_counter()
        self.pings[data] = (pong_waiter, ping_timestamp)

        await self.write_frame(True, OP_PING, data)

        return asyncio.shield(pong_waiter)

    async def read_data_frame(self, max_size: Optional[int]) -> Optional[Frame]:
        """
        Read a single data frame from the connection.

        Process control frames received before the next data frame.

        Return :obj:`None` if a close frame is encountered before any data frame.

        """
        # 6.2. Receiving Data
        while True:
            frame = await self.read_frame(max_size)

            # 5.5. Control Frames
            if frame.opcode == OP_CLOSE:
                # 7.1.5.  The WebSocket Connection Close Code
                # 7.1.6.  The WebSocket Connection Close Reason
                self.close_rcvd = Close.parse(frame.data)
                if self.close_sent is not None:
                    self.close_rcvd_then_sent = False
                try:
                    # Echo the original data instead of re-serializing it with
                    # Close.serialize() because that fails when the close frame
                    # is empty and Close.parse() synthesizes a 1005 close code.
                    await self.write_close_frame(self.close_rcvd, frame.data)
                except WebsocketException as exc:
                    # Connection closed before we could echo the close frame.
                    if exc.type == 'ConnectionClosed':
                        pass
                return None

            elif frame.opcode == OP_PING:
                # Answer pings, unless connection is CLOSING.
                if self.state is State.OPEN:
                    try:
                        await self.pong(frame.data)
                    except WebsocketException:
                        # Connection closed while draining write buffer.
                        pass

            elif frame.opcode == OP_PONG:
                if frame.data in self.pings:
                    pong_timestamp = time.perf_counter()
                    # Sending a pong for only the most recent ping is legal.
                    # Acknowledge all previous pings too in that case.
                    ping_ids = []
                    for ping_id, (pong_waiter, ping_timestamp) in self.pings.items():
                        ping_ids.append(ping_id)
                        if not pong_waiter.done():
                            pong_waiter.set_result(pong_timestamp - ping_timestamp)
                        if ping_id == frame.data:
                            self.latency = pong_timestamp - ping_timestamp
                            break
                    else:  # pragma: no cover
                        assert False, "ping_id is in self.pings"
                    # Remove acknowledged pings from self.pings.
                    for ping_id in ping_ids:
                        del self.pings[ping_id]

            # 5.6. Data Frames
            else:
                return frame

    async def read_message(self) -> Optional[Data]:
        """
        Read a single message from the connection.

        Re-assemble data frames if the message is fragmented.

        Return :obj:`None` when the closing handshake is started.

        """
        frame = await self.read_data_frame(max_size=self.max_size)

        # A close frame was received.
        if frame is None:
            return None

        if frame.opcode == OP_TEXT:
            text = True
        elif frame.opcode == OP_BINARY:
            text = False
        else:  # frame.opcode == OP_CONT
            raise WebsocketException("ProtocolError: unexpected opcode")

        # Shortcut for the common case - no fragmentation
        if frame.fin:
            return frame.data.decode("utf-8") if text else frame.data

        # 5.4. Fragmentation
        fragments: List[Data] = []
        max_size = self.max_size
        if text:
            decoder_factory = codecs.getincrementaldecoder("utf-8")
            decoder = decoder_factory(errors="strict")
            if max_size is None:

                def append(frame: Frame) -> None:
                    nonlocal fragments
                    fragments.append(decoder.decode(frame.data, frame.fin))

            else:

                def append(frame: Frame) -> None:
                    nonlocal fragments, max_size
                    fragments.append(decoder.decode(frame.data, frame.fin))
                    assert isinstance(max_size, int)
                    max_size -= len(frame.data)

        else:
            if max_size is None:

                def append(frame: Frame) -> None:
                    nonlocal fragments
                    fragments.append(frame.data)

            else:

                def append(frame: Frame) -> None:
                    nonlocal fragments, max_size
                    fragments.append(frame.data)
                    assert isinstance(max_size, int)
                    max_size -= len(frame.data)

        append(frame)

        while not frame.fin:
            frame = await self.read_data_frame(max_size=max_size)
            if frame is None:
                raise WebsocketException("ProtocolError: incomplete fragmented message")
            if frame.opcode != OP_CONT:
                raise WebsocketException("ProtocolError: unexpected opcode")
            append(frame)

        return ("" if text else b"").join(fragments)

    async def transfer_data(self) -> None:
        """
        Read incoming messages and put them in a queue.

        This coroutine runs in a task until the closing handshake is started.

        """
        try:
            while True:
                message = await self.read_message()

                # Exit the loop when receiving a close frame.
                if message is None:
                    break

                # Wait until there's room in the queue (if necessary).
                if self.max_queue is not None:
                    while len(self.messages) >= self.max_queue:
                        self._put_message_waiter = self.loop.create_future()
                        try:
                            await asyncio.shield(self._put_message_waiter)
                        finally:
                            self._put_message_waiter = None

                # Put the message in the queue.
                self.messages.append(message)

                # Notify recv().
                if self._pop_message_waiter is not None:
                    self._pop_message_waiter.set_result(None)
                    self._pop_message_waiter = None

        except asyncio.CancelledError as exc:
            self.transfer_data_exc = exc
            # If fail_connection() cancels this task, avoid logging the error
            # twice and failing the connection again.
            raise

        except WebsocketException as exc:
            if exc.type == 'ProtocolError':
                self.transfer_data_exc = exc
                self.fail_connection(1002)
            elif exc.type == 'PayloadTooBig':
                self.transfer_data_exc = exc
                self.fail_connection(1009)

        except (ConnectionError, TimeoutError, EOFError, ssl.SSLError) as exc:
            # Reading data with self.reader.readexactly may raise:
            # - most subclasses of ConnectionError if the TCP connection
            #   breaks, is reset, or is aborted;
            # - TimeoutError if the TCP connection times out;
            # - IncompleteReadError, a subclass of EOFError, if fewer
            #   bytes are available than requested;
            # - ssl.SSLError if the other side infringes the TLS protocol.
            self.transfer_data_exc = exc
            self.fail_connection(1006)

        except UnicodeDecodeError as exc:
            self.transfer_data_exc = exc
            self.fail_connection(1007)

        except Exception as exc:
            # This shouldn't happen often because exceptions expected under
            # regular circumstances are handled above. If it does, consider
            # catching and handling more exceptions.
            self.logger.error("data transfer failed", exc_info=True)

            self.transfer_data_exc = exc
            self.fail_connection(1011)

    async def keepalive_ping(self) -> None:
        """
        Send a Ping frame and wait for a Pong frame at regular intervals.

        This coroutine exits when the connection terminates and one of the
        following happens:

        - :meth:`ping` raises :exc:`ConnectionClosed`, or
        - :meth:`close_connection` cancels :attr:`keepalive_ping_task`.

        """
        if self.ping_interval is None:
            return

        try:
            while True:
                await asyncio.sleep(
                    self.ping_interval
                )

                # ping() raises CancelledError if the connection is closed,
                # when close_connection() cancels self.keepalive_ping_task.

                # ping() raises ConnectionClosed if the connection is lost,
                # when connection_lost() calls abort_pings().

                self.logger.debug("% sending keepalive ping")
                pong_waiter = await self.ping()

                if self.ping_timeout is not None:
                    try:
                        await asyncio.wait_for(
                            pong_waiter,
                            self.ping_timeout
                        )
                        self.logger.debug("% received keepalive pong")
                    except asyncio.TimeoutError:
                        if self.debug:
                            self.logger.debug("! timed out waiting for keepalive pong")
                        self.fail_connection(1011, "keepalive ping timeout")
                        break

        # Remove this branch when dropping support for Python < 3.8
        # because CancelledError no longer inherits Exception.
        except asyncio.CancelledError:
            raise

        except WebsocketException:
            pass

        except Exception:
            self.logger.error("keepalive ping failed", exc_info=True)

    def connection_open(self) -> None:
        """
        Callback when the WebSocket opening handshake completes.

        Enter the OPEN state and start the data transfer phase.

        """
        # 4.1. The WebSocket Connection is Established.
        assert self.state is State.CONNECTING
        self.state = State.OPEN
        if self.debug:
            self.logger.debug("= connection is OPEN")
        # Start the task that receives incoming WebSocket messages.
        self.transfer_data_task = self.loop.create_task(self.transfer_data())
        # Start the task that sends pings at regular intervals.
        self.keepalive_ping_task = self.loop.create_task(self.keepalive_ping())
        # Start the task that eventually closes the TCP connection.
        self.close_connection_task = self.loop.create_task(self.close_connection())

    async def handshake(
            self
    ) -> str:
        """
        Perform the server side of the opening handshake.

        Returns:
            str: path of the URI of the request.

        Raises:
            WebsocketException: if the handshake fails.

        """
        path, request_headers = await self.read_http_request()

        # The connection may drop while process_request is running.
        if self.state is State.CLOSED:
            # This subclass of ConnectionError is silently ignored in handler().
            raise BrokenPipeError("connection closed during opening handshake")

        key = check_request(request_headers)

        response_headers = Headers()

        build_response(response_headers, key)

        response_headers.setdefault("Date", email.utils.formatdate(usegmt=True))

        self.write_http_response(http.HTTPStatus.SWITCHING_PROTOCOLS, response_headers)

        self.logger.info("connection open")

        self.connection_open()

        return path

    def data_received(self, data: bytes) -> None:
        self.reader.feed_data(data)

    def abort_pings(self) -> None:
        """
        Raise ConnectionClosed in pending keepalive pings.

        They'll never receive a pong once the connection is closed.

        """
        assert self.state is State.CLOSED
        exc = self.connection_closed_exc()

        for pong_waiter, _ping_timestamp in self.pings.values():
            pong_waiter.set_exception(exc)
            # If the exception is never retrieved, it will be logged when ping
            # is garbage-collected. This is confusing for users.
            # Given that ping is done (with an exception), canceling it does
            # nothing, but it prevents logging the exception.
            pong_waiter.cancel()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """
        7.1.4. The WebSocket Connection is Closed.

        """
        self.state = State.CLOSED
        self.logger.debug("= connection is CLOSED")

        self.abort_pings()

        # If self.connection_lost_waiter isn't pending, that's a bug, because:
        # - it's set only here in connection_lost() which is called only once;
        # - it must never be canceled.
        self.connection_lost_waiter.set_result(None)

        if True:  # pragma: no cover
            # Copied from asyncio.StreamReaderProtocol
            if self.reader is not None:
                if exc is None:
                    self.reader.feed_eof()
                else:
                    self.reader.set_exception(exc)

            # Copied from asyncio.FlowControlMixin
            # Wake up the writer if currently paused.
            if not self._paused:
                return
            waiter = self._drain_waiter
            if waiter is None:
                return
            self._drain_waiter = None
            if waiter.done():
                return
            if exc is None:
                waiter.set_result(None)
            else:
                waiter.set_exception(exc)

    def pause_writing(self) -> None:
        assert not self._paused
        self._paused = True

    async def recv(self) -> Data:
        """
        Receive the next message.

        When the connection is closed, :meth:`recv` raises
        :exc:`WebsocketException`. Specifically, it raises :exc:`WebsocketException` 
        after a normal connection closure and :exc:`WebsocketException` after
        a protocol error or a network failure. This is how you detect the end of the
        message stream.

        Canceling :meth:`recv` is safe. There's no risk of losing the next
        message. The next invocation of :meth:`recv` will return it.

        This makes it possible to enforce a timeout by wrapping :meth:`recv`
        in :func:`~asyncio.wait_for`.

        Returns:
            Data: A string (:class:`str`) for a Text_ frame. A bytestring
            (:class:`bytes`) for a Binary_ frame.

            .. _Text: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.6
            .. _Binary: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.6

        Raises:
            WebsocketException: when the connection is closed.
            RuntimeError: if two coroutines call :meth:`recv` concurrently.

        """
        if self._pop_message_waiter is not None:
            raise RuntimeError(
                "cannot call recv while another coroutine "
                "is already waiting for the next message"
            )

        # Don't await self.ensure_open() here:
        # - messages could be available in the queue even if the connection
        #   is closed;
        # - messages could be received before the closing frame even if the
        #   connection is closing.

        # Wait until there's a message in the queue (if necessary) or the
        # connection is closed.
        while len(self.messages) <= 0:
            pop_message_waiter: asyncio.Future[None] = self.loop.create_future()
            self._pop_message_waiter = pop_message_waiter
            try:
                # If asyncio.wait() is canceled, it doesn't cancel
                # pop_message_waiter and self.transfer_data_task.
                await asyncio.wait(
                    [pop_message_waiter, self.transfer_data_task],
                    return_when=asyncio.FIRST_COMPLETED,
                )
            finally:
                self._pop_message_waiter = None

            # If asyncio.wait(...) exited because self.transfer_data_task
            # completed before receiving a new message, raise a suitable
            # exception (or return None if legacy_recv is enabled).
            if not pop_message_waiter.done():
                await self.ensure_open()

        # Pop a message from the queue.
        message = self.messages.popleft()

        # Notify transfer_data().
        if self._put_message_waiter is not None:
            self._put_message_waiter.set_result(None)
            self._put_message_waiter = None

        return message

    async def send(
            self,
            message: Union[Data, Iterable[Data], AsyncIterable[Data]],
    ) -> None:
        """
        Send a message.

        A string (:class:`str`) is sent as a Text_ frame. A bytestring or
        bytes-like object (:class:`bytes`, :class:`bytearray`, or
        :class:`memoryview`) is sent as a Binary_ frame.

        .. _Text: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.6
        .. _Binary: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.6

        :meth:`send` also accepts an iterable or an asynchronous iterable of
        strings, bytestrings, or bytes-like objects to enable fragmentation_.
        Each item is treated as a message fragment and sent in its own frame.
        All items must be of the same type, or else :meth:`send` will raise a
        :exc:`TypeError` and the connection will be closed.

        .. _fragmentation: https://www.rfc-editor.org/rfc/rfc6455.html#section-5.4

        :meth:`send` rejects dict-like objects because this is often an error.
        (If you want to send the keys of a dict-like object as fragments, call
        its :meth:`~dict.keys` method and pass the result to :meth:`send`.)

        Canceling :meth:`send` is discouraged. Instead, you should close the
        connection with :meth:`close`. Indeed, there are only two situations
        where :meth:`send` may yield control to the event loop and then get
        canceled; in both cases, :meth:`close` has the same effect and is
        more clear:

        1. The write buffer is full. If you don't want to wait until enough
           data is sent, your only alternative is to close the connection.
           :meth:`close` will likely time out then abort the TCP connection.
        2. ``message`` is an asynchronous iterator that yields control.
           Stopping in the middle of a fragmented message will cause a
           protocol error and the connection will be closed.

        When the connection is closed, :meth:`send` raises
        :exc:`WebsocketException`. Specifically, it raises :exc:`WebsocketException`
        after a normal connection closure and :exc:`WebsocketException` after
        a protocol error or a network failure.

        Args:
            message (Union[Data, Iterable[Data], AsyncIterable[Data]): message
                to send.

        Raises:
            WebsocketException: when the connection is closed.
            TypeError: if ``message`` doesn't have a supported type.

        """
        await self.ensure_open()

        # While sending a fragmented message, prevent sending other messages
        # until all fragments are sent.
        while self._fragmented_message_waiter is not None:
            await asyncio.shield(self._fragmented_message_waiter)

        # Unfragmented message -- this case must be handled first because
        # strings and bytes-like objects are iterable.

        if isinstance(message, (str, bytes, bytearray, memoryview)):
            opcode, data = prepare_data(message)
            await self.write_frame(True, opcode, data)

        # Catch a common mistake -- passing a dict to send().

        elif isinstance(message, Mapping):
            raise TypeError("data is a dict-like object")

        # Fragmented message -- regular iterator.

        elif isinstance(message, Iterable):

            # Work around https://github.com/python/mypy/issues/6227
            message = cast(Iterable[Data], message)

            iter_message = iter(message)
            try:
                fragment = next(iter_message)
            except StopIteration:
                return
            opcode, data = prepare_data(fragment)

            self._fragmented_message_waiter = asyncio.Future()
            try:
                # First fragment.
                await self.write_frame(False, opcode, data)

                # Other fragments.
                for fragment in iter_message:
                    confirm_opcode, data = prepare_data(fragment)
                    if confirm_opcode != opcode:
                        raise TypeError("data contains inconsistent types")
                    await self.write_frame(False, OP_CONT, data)

                # Final fragment.
                await self.write_frame(True, OP_CONT, b"")

            except (Exception, asyncio.CancelledError):
                # We're half-way through a fragmented message and we can't
                # complete it. This makes the connection unusable.
                self.fail_connection(1011)
                raise

            finally:
                self._fragmented_message_waiter.set_result(None)
                self._fragmented_message_waiter = None

        # Fragmented message -- asynchronous iterator

        elif isinstance(message, AsyncIterable):
            # Implement aiter_message = aiter(message) without aiter
            # Work around https://github.com/python/mypy/issues/5738
            aiter_message = cast(
                Callable[[AsyncIterable[Data]], AsyncIterator[Data]],
                type(message).__aiter__,
            )(message)
            try:
                # Implement fragment = anext(aiter_message) without anext
                # Work around https://github.com/python/mypy/issues/5738
                fragment = await cast(
                    Callable[[AsyncIterator[Data]], Awaitable[Data]],
                    type(aiter_message).__anext__,
                )(aiter_message)
            except StopAsyncIteration:
                return
            opcode, data = prepare_data(fragment)

            self._fragmented_message_waiter = asyncio.Future()
            try:
                # First fragment.
                await self.write_frame(False, opcode, data)

                # Other fragments.
                # coverage reports this code as not covered, but it is
                # exercised by tests - changing it breaks the tests!
                async for fragment in aiter_message:  # pragma: no cover
                    confirm_opcode, data = prepare_data(fragment)
                    if confirm_opcode != opcode:
                        raise TypeError("data contains inconsistent types")
                    await self.write_frame(False, OP_CONT, data)

                # Final fragment.
                await self.write_frame(True, OP_CONT, b"")

            except (Exception, asyncio.CancelledError):
                # We're half-way through a fragmented message and we can't
                # complete it. This makes the connection unusable.
                self.fail_connection(1011)
                raise

            finally:
                self._fragmented_message_waiter.set_result(None)
                self._fragmented_message_waiter = None

        else:
            raise TypeError("data must be str, bytes-like, or iterable")

    async def __aiter__(self) -> AsyncIterator[Data]:
        """
        Iterate on incoming messages.

        The iterator  exits normally when the connection is closed with the
        close code 1000 (OK) or 1001(going away) or without a close code. It
        raises a :exc:`~websockets.exceptions.ConnectionClosedError` exception
        when the connection is closed with any other code.

        """
        try:
            while True:
                yield await self.recv()
        except RuntimeError:
            return
